#AUTOGENERATED!!! Date: 2020-06-19 18:55:28.807108

from enum import IntEnum

class AttributeIds(IntEnum):
    NodeId = 1
    NodeClass = 2
    BrowseName = 3
    DisplayName = 4
    Description = 5
    WriteMask = 6
    UserWriteMask = 7
    IsAbstract = 8
    Symmetric = 9
    InverseName = 10
    ContainsNoLoops = 11
    EventNotifier = 12
    Value = 13
    DataType = 14
    ValueRank = 15
    ArrayDimensions = 16
    AccessLevel = 17
    UserAccessLevel = 18
    MinimumSamplingInterval = 19
    Historizing = 20
    Executable = 21
    UserExecutable = 22
    DataTypeDefinition = 23
    RolePermissions = 24
    UserRolePermissions = 25
    AccessRestrictions = 26
    AccessLevelEx = 27



import logging

from opcua import ua
from opcua.common.node import Node


logger = logging.getLogger(__name__)


def copy_node(parent, node, nodeid=None, recursive=True):
    """
    Copy a node or node tree as child of parent node
    """
    rdesc = _rdesc_from_node(parent, node)

    if nodeid is None:
        nodeid = ua.NodeId(namespaceidx=node.nodeid.NamespaceIndex)
    added_nodeids = _copy_node(parent.server, parent.nodeid, rdesc, nodeid, recursive)
    return [Node(parent.server, nid) for nid in added_nodeids]


def _copy_node(server, parent_nodeid, rdesc, nodeid, recursive):
    addnode = ua.AddNodesItem()
    addnode.RequestedNewNodeId = nodeid
    addnode.BrowseName = rdesc.BrowseName
    addnode.ParentNodeId = parent_nodeid
    addnode.ReferenceTypeId = rdesc.ReferenceTypeId
    addnode.TypeDefinition = rdesc.TypeDefinition
    addnode.NodeClass = rdesc.NodeClass

    node_to_copy = Node(server, rdesc.NodeId)
    
    attrObj = getattr(ua, rdesc.NodeClass.name + "Attributes")
    _read_and_copy_attrs(node_to_copy, attrObj(), addnode)
    
    res = server.add_nodes([addnode])[0]

    added_nodes = [res.AddedNodeId]
        
    if recursive:
        descs = node_to_copy.get_children_descriptions()
        for desc in descs:
            nodes = _copy_node(server, res.AddedNodeId, desc, nodeid=ua.NodeId(namespaceidx=desc.NodeId.NamespaceIndex), recursive=True)
            added_nodes.extend(nodes)

    return added_nodes


def _rdesc_from_node(parent, node):
    results = node.get_attributes([ua.AttributeIds.NodeClass, ua.AttributeIds.BrowseName, ua.AttributeIds.DisplayName])
    nclass, qname, dname = [res.Value.Value for res in results]

    rdesc = ua.ReferenceDescription()
    rdesc.NodeId = node.nodeid
    rdesc.BrowseName = qname
    rdesc.DisplayName = dname
    rdesc.NodeClass = nclass
    if parent.get_type_definition() == ua.NodeId(ua.ObjectIds.FolderType):
        rdesc.ReferenceTypeId = ua.NodeId(ua.ObjectIds.Organizes)
    else:
        rdesc.ReferenceTypeId = ua.NodeId(ua.ObjectIds.HasComponent)
    typedef = node.get_type_definition()
    if typedef:
        rdesc.TypeDefinition = typedef
    return rdesc


def _read_and_copy_attrs(node_type, struct, addnode):
    names = [name for name in struct.__dict__.keys() if not name.startswith("_") and name not in ("BodyLength", "TypeId", "SpecifiedAttributes", "Encoding", "IsAbstract", "EventNotifier")]
    attrs = [getattr(ua.AttributeIds, name) for name in names]            
    for name in names:
        results = node_type.get_attributes(attrs)
    for idx, name in enumerate(names):
        if results[idx].StatusCode.is_good():
            if name == "Value":
                setattr(struct, name, results[idx].Value)
            else:
                setattr(struct, name, results[idx].Value.Value)
        else:
            logger.warning("Instantiate: while copying attributes from node type {0!s}, attribute {1!s}, statuscode is {2!s}".format(node_type, name, results[idx].StatusCode))            
    addnode.NodeAttributes = struct



#AUTOGENERATED!!! Date: 2020-06-19 19:47:20.401119

from opcua.ua.uaerrors import UaStatusCodeError


class Bad(UaStatusCodeError):
    code = 0x80000000

class BadUnexpectedError(UaStatusCodeError):
    code = 0x80010000

class BadInternalError(UaStatusCodeError):
    code = 0x80020000

class BadOutOfMemory(UaStatusCodeError):
    code = 0x80030000

class BadResourceUnavailable(UaStatusCodeError):
    code = 0x80040000

class BadCommunicationError(UaStatusCodeError):
    code = 0x80050000

class BadEncodingError(UaStatusCodeError):
    code = 0x80060000

class BadDecodingError(UaStatusCodeError):
    code = 0x80070000

class BadEncodingLimitsExceeded(UaStatusCodeError):
    code = 0x80080000

class BadRequestTooLarge(UaStatusCodeError):
    code = 0x80B80000

class BadResponseTooLarge(UaStatusCodeError):
    code = 0x80B90000

class BadUnknownResponse(UaStatusCodeError):
    code = 0x80090000

class BadTimeout(UaStatusCodeError):
    code = 0x800A0000

class BadServiceUnsupported(UaStatusCodeError):
    code = 0x800B0000

class BadShutdown(UaStatusCodeError):
    code = 0x800C0000

class BadServerNotConnected(UaStatusCodeError):
    code = 0x800D0000

class BadServerHalted(UaStatusCodeError):
    code = 0x800E0000

class BadNothingToDo(UaStatusCodeError):
    code = 0x800F0000

class BadTooManyOperations(UaStatusCodeError):
    code = 0x80100000

class BadTooManyMonitoredItems(UaStatusCodeError):
    code = 0x80DB0000

class BadDataTypeIdUnknown(UaStatusCodeError):
    code = 0x80110000

class BadCertificateInvalid(UaStatusCodeError):
    code = 0x80120000

class BadSecurityChecksFailed(UaStatusCodeError):
    code = 0x80130000

class BadCertificatePolicyCheckFailed(UaStatusCodeError):
    code = 0x81140000

class BadCertificateTimeInvalid(UaStatusCodeError):
    code = 0x80140000

class BadCertificateIssuerTimeInvalid(UaStatusCodeError):
    code = 0x80150000

class BadCertificateHostNameInvalid(UaStatusCodeError):
    code = 0x80160000

class BadCertificateUriInvalid(UaStatusCodeError):
    code = 0x80170000

class BadCertificateUseNotAllowed(UaStatusCodeError):
    code = 0x80180000

class BadCertificateIssuerUseNotAllowed(UaStatusCodeError):
    code = 0x80190000

class BadCertificateUntrusted(UaStatusCodeError):
    code = 0x801A0000

class BadCertificateRevocationUnknown(UaStatusCodeError):
    code = 0x801B0000

class BadCertificateIssuerRevocationUnknown(UaStatusCodeError):
    code = 0x801C0000

class BadCertificateRevoked(UaStatusCodeError):
    code = 0x801D0000

class BadCertificateIssuerRevoked(UaStatusCodeError):
    code = 0x801E0000

class BadCertificateChainIncomplete(UaStatusCodeError):
    code = 0x810D0000

class BadUserAccessDenied(UaStatusCodeError):
    code = 0x801F0000

class BadIdentityTokenInvalid(UaStatusCodeError):
    code = 0x80200000

class BadIdentityTokenRejected(UaStatusCodeError):
    code = 0x80210000

class BadSecureChannelIdInvalid(UaStatusCodeError):
    code = 0x80220000

class BadInvalidTimestamp(UaStatusCodeError):
    code = 0x80230000

class BadNonceInvalid(UaStatusCodeError):
    code = 0x80240000

class BadSessionIdInvalid(UaStatusCodeError):
    code = 0x80250000

class BadSessionClosed(UaStatusCodeError):
    code = 0x80260000

class BadSessionNotActivated(UaStatusCodeError):
    code = 0x80270000

class BadSubscriptionIdInvalid(UaStatusCodeError):
    code = 0x80280000

class BadRequestHeaderInvalid(UaStatusCodeError):
    code = 0x802A0000

class BadTimestampsToReturnInvalid(UaStatusCodeError):
    code = 0x802B0000

class BadRequestCancelledByClient(UaStatusCodeError):
    code = 0x802C0000

class BadTooManyArguments(UaStatusCodeError):
    code = 0x80E50000

class BadLicenseExpired(UaStatusCodeError):
    code = 0x810E0000

class BadLicenseLimitsExceeded(UaStatusCodeError):
    code = 0x810F0000

class BadLicenseNotAvailable(UaStatusCodeError):
    code = 0x81100000

class BadNoCommunication(UaStatusCodeError):
    code = 0x80310000

class BadWaitingForInitialData(UaStatusCodeError):
    code = 0x80320000

class BadNodeIdInvalid(UaStatusCodeError):
    code = 0x80330000

class BadNodeIdUnknown(UaStatusCodeError):
    code = 0x80340000

class BadAttributeIdInvalid(UaStatusCodeError):
    code = 0x80350000

class BadIndexRangeInvalid(UaStatusCodeError):
    code = 0x80360000

class BadIndexRangeNoData(UaStatusCodeError):
    code = 0x80370000

class BadDataEncodingInvalid(UaStatusCodeError):
    code = 0x80380000

class BadDataEncodingUnsupported(UaStatusCodeError):
    code = 0x80390000

class BadNotReadable(UaStatusCodeError):
    code = 0x803A0000

class BadNotWritable(UaStatusCodeError):
    code = 0x803B0000

class BadOutOfRange(UaStatusCodeError):
    code = 0x803C0000

class BadNotSupported(UaStatusCodeError):
    code = 0x803D0000

class BadNotFound(UaStatusCodeError):
    code = 0x803E0000

class BadObjectDeleted(UaStatusCodeError):
    code = 0x803F0000

class BadNotImplemented(UaStatusCodeError):
    code = 0x80400000

class BadMonitoringModeInvalid(UaStatusCodeError):
    code = 0x80410000

class BadMonitoredItemIdInvalid(UaStatusCodeError):
    code = 0x80420000

class BadMonitoredItemFilterInvalid(UaStatusCodeError):
    code = 0x80430000

class BadMonitoredItemFilterUnsupported(UaStatusCodeError):
    code = 0x80440000

class BadFilterNotAllowed(UaStatusCodeError):
    code = 0x80450000

class BadStructureMissing(UaStatusCodeError):
    code = 0x80460000

class BadEventFilterInvalid(UaStatusCodeError):
    code = 0x80470000

class BadContentFilterInvalid(UaStatusCodeError):
    code = 0x80480000

class BadFilterOperatorInvalid(UaStatusCodeError):
    code = 0x80C10000

class BadFilterOperatorUnsupported(UaStatusCodeError):
    code = 0x80C20000

class BadFilterOperandCountMismatch(UaStatusCodeError):
    code = 0x80C30000

class BadFilterOperandInvalid(UaStatusCodeError):
    code = 0x80490000

class BadFilterElementInvalid(UaStatusCodeError):
    code = 0x80C40000

class BadFilterLiteralInvalid(UaStatusCodeError):
    code = 0x80C50000

class BadContinuationPointInvalid(UaStatusCodeError):
    code = 0x804A0000

class BadNoContinuationPoints(UaStatusCodeError):
    code = 0x804B0000

class BadReferenceTypeIdInvalid(UaStatusCodeError):
    code = 0x804C0000

class BadBrowseDirectionInvalid(UaStatusCodeError):
    code = 0x804D0000

class BadNodeNotInView(UaStatusCodeError):
    code = 0x804E0000

class BadNumericOverflow(UaStatusCodeError):
    code = 0x81120000

class BadServerUriInvalid(UaStatusCodeError):
    code = 0x804F0000

class BadServerNameMissing(UaStatusCodeError):
    code = 0x80500000

class BadDiscoveryUrlMissing(UaStatusCodeError):
    code = 0x80510000

class BadSempahoreFileMissing(UaStatusCodeError):
    code = 0x80520000

class BadRequestTypeInvalid(UaStatusCodeError):
    code = 0x80530000

class BadSecurityModeRejected(UaStatusCodeError):
    code = 0x80540000

class BadSecurityPolicyRejected(UaStatusCodeError):
    code = 0x80550000

class BadTooManySessions(UaStatusCodeError):
    code = 0x80560000

class BadUserSignatureInvalid(UaStatusCodeError):
    code = 0x80570000

class BadApplicationSignatureInvalid(UaStatusCodeError):
    code = 0x80580000

class BadNoValidCertificates(UaStatusCodeError):
    code = 0x80590000

class BadIdentityChangeNotSupported(UaStatusCodeError):
    code = 0x80C60000

class BadRequestCancelledByRequest(UaStatusCodeError):
    code = 0x805A0000

class BadParentNodeIdInvalid(UaStatusCodeError):
    code = 0x805B0000

class BadReferenceNotAllowed(UaStatusCodeError):
    code = 0x805C0000

class BadNodeIdRejected(UaStatusCodeError):
    code = 0x805D0000

class BadNodeIdExists(UaStatusCodeError):
    code = 0x805E0000

class BadNodeClassInvalid(UaStatusCodeError):
    code = 0x805F0000

class BadBrowseNameInvalid(UaStatusCodeError):
    code = 0x80600000

class BadBrowseNameDuplicated(UaStatusCodeError):
    code = 0x80610000

class BadNodeAttributesInvalid(UaStatusCodeError):
    code = 0x80620000

class BadTypeDefinitionInvalid(UaStatusCodeError):
    code = 0x80630000

class BadSourceNodeIdInvalid(UaStatusCodeError):
    code = 0x80640000

class BadTargetNodeIdInvalid(UaStatusCodeError):
    code = 0x80650000

class BadDuplicateReferenceNotAllowed(UaStatusCodeError):
    code = 0x80660000

class BadInvalidSelfReference(UaStatusCodeError):
    code = 0x80670000

class BadReferenceLocalOnly(UaStatusCodeError):
    code = 0x80680000

class BadNoDeleteRights(UaStatusCodeError):
    code = 0x80690000

class BadServerIndexInvalid(UaStatusCodeError):
    code = 0x806A0000

class BadViewIdUnknown(UaStatusCodeError):
    code = 0x806B0000

class BadViewTimestampInvalid(UaStatusCodeError):
    code = 0x80C90000

class BadViewParameterMismatch(UaStatusCodeError):
    code = 0x80CA0000

class BadViewVersionInvalid(UaStatusCodeError):
    code = 0x80CB0000

class BadNotTypeDefinition(UaStatusCodeError):
    code = 0x80C80000

class BadTooManyMatches(UaStatusCodeError):
    code = 0x806D0000

class BadQueryTooComplex(UaStatusCodeError):
    code = 0x806E0000

class BadNoMatch(UaStatusCodeError):
    code = 0x806F0000

class BadMaxAgeInvalid(UaStatusCodeError):
    code = 0x80700000

class BadSecurityModeInsufficient(UaStatusCodeError):
    code = 0x80E60000

class BadHistoryOperationInvalid(UaStatusCodeError):
    code = 0x80710000

class BadHistoryOperationUnsupported(UaStatusCodeError):
    code = 0x80720000

class BadInvalidTimestampArgument(UaStatusCodeError):
    code = 0x80BD0000

class BadWriteNotSupported(UaStatusCodeError):
    code = 0x80730000

class BadTypeMismatch(UaStatusCodeError):
    code = 0x80740000

class BadMethodInvalid(UaStatusCodeError):
    code = 0x80750000

class BadArgumentsMissing(UaStatusCodeError):
    code = 0x80760000

class BadNotExecutable(UaStatusCodeError):
    code = 0x81110000

class BadTooManySubscriptions(UaStatusCodeError):
    code = 0x80770000

class BadTooManyPublishRequests(UaStatusCodeError):
    code = 0x80780000

class BadNoSubscription(UaStatusCodeError):
    code = 0x80790000

class BadSequenceNumberUnknown(UaStatusCodeError):
    code = 0x807A0000

class BadMessageNotAvailable(UaStatusCodeError):
    code = 0x807B0000

class BadInsufficientClientProfile(UaStatusCodeError):
    code = 0x807C0000

class BadStateNotActive(UaStatusCodeError):
    code = 0x80BF0000

class BadAlreadyExists(UaStatusCodeError):
    code = 0x81150000

class BadTcpServerTooBusy(UaStatusCodeError):
    code = 0x807D0000

class BadTcpMessageTypeInvalid(UaStatusCodeError):
    code = 0x807E0000

class BadTcpSecureChannelUnknown(UaStatusCodeError):
    code = 0x807F0000

class BadTcpMessageTooLarge(UaStatusCodeError):
    code = 0x80800000

class BadTcpNotEnoughResources(UaStatusCodeError):
    code = 0x80810000

class BadTcpInternalError(UaStatusCodeError):
    code = 0x80820000

class BadTcpEndpointUrlInvalid(UaStatusCodeError):
    code = 0x80830000

class BadRequestInterrupted(UaStatusCodeError):
    code = 0x80840000

class BadRequestTimeout(UaStatusCodeError):
    code = 0x80850000

class BadSecureChannelClosed(UaStatusCodeError):
    code = 0x80860000

class BadSecureChannelTokenUnknown(UaStatusCodeError):
    code = 0x80870000

class BadSequenceNumberInvalid(UaStatusCodeError):
    code = 0x80880000

class BadProtocolVersionUnsupported(UaStatusCodeError):
    code = 0x80BE0000

class BadConfigurationError(UaStatusCodeError):
    code = 0x80890000

class BadNotConnected(UaStatusCodeError):
    code = 0x808A0000

class BadDeviceFailure(UaStatusCodeError):
    code = 0x808B0000

class BadSensorFailure(UaStatusCodeError):
    code = 0x808C0000

class BadOutOfService(UaStatusCodeError):
    code = 0x808D0000

class BadDeadbandFilterInvalid(UaStatusCodeError):
    code = 0x808E0000

class BadRefreshInProgress(UaStatusCodeError):
    code = 0x80970000

class BadConditionAlreadyDisabled(UaStatusCodeError):
    code = 0x80980000

class BadConditionAlreadyEnabled(UaStatusCodeError):
    code = 0x80CC0000

class BadConditionDisabled(UaStatusCodeError):
    code = 0x80990000

class BadEventIdUnknown(UaStatusCodeError):
    code = 0x809A0000

class BadEventNotAcknowledgeable(UaStatusCodeError):
    code = 0x80BB0000

class BadDialogNotActive(UaStatusCodeError):
    code = 0x80CD0000

class BadDialogResponseInvalid(UaStatusCodeError):
    code = 0x80CE0000

class BadConditionBranchAlreadyAcked(UaStatusCodeError):
    code = 0x80CF0000

class BadConditionBranchAlreadyConfirmed(UaStatusCodeError):
    code = 0x80D00000

class BadConditionAlreadyShelved(UaStatusCodeError):
    code = 0x80D10000

class BadConditionNotShelved(UaStatusCodeError):
    code = 0x80D20000

class BadShelvingTimeOutOfRange(UaStatusCodeError):
    code = 0x80D30000

class BadNoData(UaStatusCodeError):
    code = 0x809B0000

class BadBoundNotFound(UaStatusCodeError):
    code = 0x80D70000

class BadBoundNotSupported(UaStatusCodeError):
    code = 0x80D80000

class BadDataLost(UaStatusCodeError):
    code = 0x809D0000

class BadDataUnavailable(UaStatusCodeError):
    code = 0x809E0000

class BadEntryExists(UaStatusCodeError):
    code = 0x809F0000

class BadNoEntryExists(UaStatusCodeError):
    code = 0x80A00000

class BadTimestampNotSupported(UaStatusCodeError):
    code = 0x80A10000

class BadAggregateListMismatch(UaStatusCodeError):
    code = 0x80D40000

class BadAggregateNotSupported(UaStatusCodeError):
    code = 0x80D50000

class BadAggregateInvalidInputs(UaStatusCodeError):
    code = 0x80D60000

class BadAggregateConfigurationRejected(UaStatusCodeError):
    code = 0x80DA0000

class BadRequestNotAllowed(UaStatusCodeError):
    code = 0x80E40000

class BadRequestNotComplete(UaStatusCodeError):
    code = 0x81130000

class BadDominantValueChanged(UaStatusCodeError):
    code = 0x80E10000

class BadDependentValueChanged(UaStatusCodeError):
    code = 0x80E30000

class BadEdited_OutOfRange(UaStatusCodeError):
    code = 0x81190000

class BadInitialValue_OutOfRange(UaStatusCodeError):
    code = 0x811A0000

class BadOutOfRange_DominantValueChanged(UaStatusCodeError):
    code = 0x811B0000

class BadEdited_OutOfRange_DominantValueChanged(UaStatusCodeError):
    code = 0x811C0000

class BadOutOfRange_DominantValueChanged_DependentValueChanged(UaStatusCodeError):
    code = 0x811D0000

class BadEdited_OutOfRange_DominantValueChanged_DependentValueChanged(UaStatusCodeError):
    code = 0x811E0000

class BadInvalidArgument(UaStatusCodeError):
    code = 0x80AB0000

class BadConnectionRejected(UaStatusCodeError):
    code = 0x80AC0000

class BadDisconnect(UaStatusCodeError):
    code = 0x80AD0000

class BadConnectionClosed(UaStatusCodeError):
    code = 0x80AE0000

class BadInvalidState(UaStatusCodeError):
    code = 0x80AF0000

class BadEndOfStream(UaStatusCodeError):
    code = 0x80B00000

class BadNoDataAvailable(UaStatusCodeError):
    code = 0x80B10000

class BadWaitingForResponse(UaStatusCodeError):
    code = 0x80B20000

class BadOperationAbandoned(UaStatusCodeError):
    code = 0x80B30000

class BadExpectedStreamToBlock(UaStatusCodeError):
    code = 0x80B40000

class BadWouldBlock(UaStatusCodeError):
    code = 0x80B50000

class BadSyntaxError(UaStatusCodeError):
    code = 0x80B60000

class BadMaxConnectionsReached(UaStatusCodeError):
    code = 0x80B70000




import os

from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import hmac
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import modes


def load_certificate(path):
    _, ext = os.path.splitext(path)
    with open(path, "rb") as f:
        if ext == ".pem":
            return x509.load_pem_x509_certificate(f.read(), default_backend())
        else:
            return x509.load_der_x509_certificate(f.read(), default_backend())


def x509_from_der(data):
    if not data:
        return None
    return x509.load_der_x509_certificate(data, default_backend())


def load_private_key(path):
    _, ext = os.path.splitext(path)
    with open(path, "rb") as f:
        if ext == ".pem":
            return serialization.load_pem_private_key(f.read(), password=None, backend=default_backend())
        else:
            return serialization.load_der_private_key(f.read(), password=None, backend=default_backend())


def der_from_x509(certificate):
    if certificate is None:
        return b""
    return certificate.public_bytes(serialization.Encoding.DER)


def sign_sha1(private_key, data):
    return private_key.sign(
        data,
        padding.PKCS1v15(),
        hashes.SHA1()
    )

def sign_sha256(private_key, data):
    return private_key.sign(
        data,
        padding.PKCS1v15(),
        hashes.SHA256()
    )

def verify_sha1(certificate, data, signature):
    certificate.public_key().verify(
        signature,
        data,
        padding.PKCS1v15(),
        hashes.SHA1())

def verify_sha256(certificate, data, signature):
    certificate.public_key().verify(
        signature,
        data,
        padding.PKCS1v15(),
        hashes.SHA256())

def encrypt_basic256(public_key, data):
    ciphertext = public_key.encrypt(
        data,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None)
    )
    return ciphertext


def encrypt_rsa_oaep(public_key, data):
    ciphertext = public_key.encrypt(
        data,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None)
    )
    return ciphertext


def encrypt_rsa15(public_key, data):
    ciphertext = public_key.encrypt(
        data,
        padding.PKCS1v15()
    )
    return ciphertext


def decrypt_rsa_oaep(private_key, data):
    text = private_key.decrypt(
        data,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None)
    )
    return text


def decrypt_rsa15(private_key, data):
    text = private_key.decrypt(
        data,
        padding.PKCS1v15()
    )
    return text


def cipher_aes_cbc(key, init_vec):
    return Cipher(algorithms.AES(key), modes.CBC(init_vec), default_backend())


def cipher_encrypt(cipher, data):
    encryptor = cipher.encryptor()
    return encryptor.update(data) + encryptor.finalize()


def cipher_decrypt(cipher, data):
    decryptor = cipher.decryptor()
    return decryptor.update(data) + decryptor.finalize()


def hmac_sha1(key, message):
    hasher = hmac.HMAC(key, hashes.SHA1(), backend=default_backend())
    hasher.update(message)
    return hasher.finalize()

def hmac_sha256(key, message):
    hasher = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
    hasher.update(message)
    return hasher.finalize()

def sha1_size():
    return hashes.SHA1.digest_size

def sha256_size():
    return hashes.SHA256.digest_size

def p_sha1(secret, seed, sizes=()):
    """
    Derive one or more keys from secret and seed.
    (See specs part 6, 6.7.5 and RFC 2246 - TLS v1.0)
    Lengths of keys will match sizes argument
    """
    full_size = 0
    for size in sizes:
        full_size += size

    result = b''
    accum = seed
    while len(result) < full_size:
        accum = hmac_sha1(secret, accum)
        result += hmac_sha1(secret, accum + seed)

    parts = []
    for size in sizes:
        parts.append(result[:size])
        result = result[size:]
    return tuple(parts)

def p_sha256(secret, seed, sizes=()):
    """
    Derive one or more keys from secret and seed.
    (See specs part 6, 6.7.5 and RFC 2246 - TLS v1.0)
    Lengths of keys will match sizes argument
    """
    full_size = 0
    for size in sizes:
        full_size += size

    result = b''
    accum = seed
    while len(result) < full_size:
        accum = hmac_sha256(secret, accum)
        result += hmac_sha256(secret, accum + seed)

    parts = []
    for size in sizes:
        parts.append(result[:size])
        result = result[size:]
    return tuple(parts)

def x509_name_to_string(name):
    parts = ["{0}={1}".format(attr.oid._name, attr.value) for attr in name]
    return ', '.join(parts)


def x509_to_string(cert):
    """
    Convert x509 certificate to human-readable string
    """
    if cert.subject == cert.issuer:
        issuer = ' (self-signed)'
    else:
        issuer = ', issuer: {0}'.format(x509_name_to_string(cert.issuer))
    # TODO: show more information
    return "{0}{1}, {2} - {3}".format(x509_name_to_string(cert.subject), issuer, cert.not_valid_before, cert.not_valid_after)


if __name__ == "__main__":
    # Convert from PEM to DER
    cert = load_certificate("../examples/server_cert.pem")
    #rsa_pubkey = pubkey_from_dercert(der)
    rsa_privkey = load_private_key("../examples/mykey.pem")

    from IPython import embed
    embed()




import logging
from threading import RLock, Lock
import time

from opcua import ua
from opcua.ua.ua_binary import nodeid_from_binary, struct_from_binary
from opcua.ua.ua_binary import struct_to_binary, uatcp_to_binary
from opcua.common import utils
from opcua.common.connection import SecureConnection


class PublishRequestData(object):

    def __init__(self):
        self.requesthdr = None
        self.seqhdr = None
        self.timestamp = time.time()


class UaProcessor(object):

    def __init__(self, internal_server, socket):
        self.logger = logging.getLogger(__name__)
        self.iserver = internal_server
        self.name = socket.get_extra_info('peername')
        self.sockname = socket.get_extra_info('sockname')
        self.session = None
        self.socket = socket
        self._socketlock = Lock()
        self._datalock = RLock()
        self._publishdata_queue = []
        self._publish_result_queue = []  # used when we need to wait for PublishRequest
        self._connection = SecureConnection(ua.SecurityPolicy())

    @property
    def local_discovery_service(self):
        return self.iserver.local_discovery_service

    def set_policies(self, policies):
        self._connection.set_policy_factories(policies)

    def send_response(self, requesthandle, seqhdr, response, msgtype=ua.MessageType.SecureMessage):
        with self._socketlock:
            response.ResponseHeader.RequestHandle = requesthandle
            data = self._connection.message_to_binary(
                struct_to_binary(response), message_type=msgtype, request_id=seqhdr.RequestId)

            self.socket.write(data)

    def open_secure_channel(self, algohdr, seqhdr, body):
        request = struct_from_binary(ua.OpenSecureChannelRequest, body)

        if not self._connection.is_open():
            # Only call select_policy if the channel isn't open. Otherwise
            # it will break the Secure channel renewal.
            self._connection.select_policy(
                algohdr.SecurityPolicyURI, algohdr.SenderCertificate, request.Parameters.SecurityMode)

        channel = self._connection.open(request.Parameters, self.iserver)
        # send response
        response = ua.OpenSecureChannelResponse()
        response.Parameters = channel
        self.send_response(request.RequestHeader.RequestHandle, seqhdr, response, ua.MessageType.SecureOpen)

    def forward_publish_response(self, result):
        self.logger.info("forward publish response %s", result)
        with self._datalock:
            while True:
                if len(self._publishdata_queue) == 0:
                    self._publish_result_queue.append(result)
                    self.logger.info("Server wants to send publish answer but no publish request is available,"
                                     "enqueing notification, length of result queue is %s",
                                     len(self._publish_result_queue))
                    return
                requestdata = self._publishdata_queue.pop(0)
                if requestdata.requesthdr.TimeoutHint == 0 or requestdata.requesthdr.TimeoutHint != 0 and time.time() - requestdata.timestamp < requestdata.requesthdr.TimeoutHint / 1000:
                    break

        response = ua.PublishResponse()
        response.Parameters = result

        self.send_response(requestdata.requesthdr.RequestHandle, requestdata.seqhdr, response)

    def process(self, header, body):
        msg = self._connection.receive_from_header_and_body(header, body)
        if isinstance(msg, ua.Message):
            if header.MessageType == ua.MessageType.SecureOpen:
                self.open_secure_channel(msg.SecurityHeader(), msg.SequenceHeader(), msg.body())

            elif header.MessageType == ua.MessageType.SecureClose:
                self._connection.close()
                return False

            elif header.MessageType == ua.MessageType.SecureMessage:
                return self.process_message(msg.SequenceHeader(), msg.body())
        elif isinstance(msg, ua.Hello):
            ack = ua.Acknowledge()
            ack.ReceiveBufferSize = msg.ReceiveBufferSize
            ack.SendBufferSize = msg.SendBufferSize
            data = uatcp_to_binary(ua.MessageType.Acknowledge, ack)
            self.socket.write(data)
        elif isinstance(msg, ua.ErrorMessage):
            self.logger.warning("Received an error message type")
        elif msg is None:
            pass  # msg is a ChunkType.Intermediate of an ua.MessageType.SecureMessage
        else:
            self.logger.warning("Unsupported message type: %s", header.MessageType)
            raise utils.ServiceError(ua.StatusCodes.BadTcpMessageTypeInvalid)
        return True

    def process_message(self, seqhdr, body):
        typeid = nodeid_from_binary(body)
        requesthdr = struct_from_binary(ua.RequestHeader, body)
        try:
            return self._process_message(typeid, requesthdr, seqhdr, body)
        except utils.ServiceError as e:
            status = ua.StatusCode(e.code)
            response = ua.ServiceFault()
            response.ResponseHeader.ServiceResult = status
            self.logger.info("sending service fault response: %s (%s)", status.doc, status.name)
            self.send_response(requesthdr.RequestHandle, seqhdr, response)
            return True

    def _process_message(self, typeid, requesthdr, seqhdr, body):
        if typeid == ua.NodeId(ua.ObjectIds.CreateSessionRequest_Encoding_DefaultBinary):
            self.logger.info("Create session request")
            params = struct_from_binary(ua.CreateSessionParameters, body)

            # create the session on server
            self.session = self.iserver.create_session(self.name)
            # get a session creation result to send back
            sessiondata = self.session.create_session(params, sockname=self.sockname)

            response = ua.CreateSessionResponse()
            response.Parameters = sessiondata
            response.Parameters.ServerCertificate = self._connection.security_policy.client_certificate
            if self._connection.security_policy.server_certificate is None:
                data = params.ClientNonce
            else:
                data = self._connection.security_policy.server_certificate + params.ClientNonce
            response.Parameters.ServerSignature.Signature = \
                self._connection.security_policy.asymmetric_cryptography.signature(data)

            response.Parameters.ServerSignature.Algorithm = self._connection.security_policy.AsymmetricSignatureURI

            self.logger.info("sending create session response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.CloseSessionRequest_Encoding_DefaultBinary):
            self.logger.info("Close session request")

            if self.session:
                deletesubs = ua.ua_binary.Primitives.Boolean.unpack(body)
                self.session.close_session(deletesubs)
            else:
                self.logger.info("Request to close non-existing session")

            response = ua.CloseSessionResponse()
            self.logger.info("sending close session response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.ActivateSessionRequest_Encoding_DefaultBinary):
            self.logger.info("Activate session request")
            params = struct_from_binary(ua.ActivateSessionParameters, body)

            if not self.session:
                self.logger.info("request to activate non-existing session")
                raise utils.ServiceError(ua.StatusCodes.BadSessionIdInvalid)

            if self._connection.security_policy.client_certificate is None:
                data = self.session.nonce
            else:
                data = self._connection.security_policy.client_certificate + self.session.nonce
            self._connection.security_policy.asymmetric_cryptography.verify(data, params.ClientSignature.Signature)

            result = self.session.activate_session(params)

            response = ua.ActivateSessionResponse()
            response.Parameters = result

            self.logger.info("sending read response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.ReadRequest_Encoding_DefaultBinary):
            self.logger.info("Read request")
            params = struct_from_binary(ua.ReadParameters, body)

            results = self.session.read(params)

            response = ua.ReadResponse()
            response.Results = results

            self.logger.info("sending read response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.WriteRequest_Encoding_DefaultBinary):
            self.logger.info("Write request")
            params = struct_from_binary(ua.WriteParameters, body)

            results = self.session.write(params)

            response = ua.WriteResponse()
            response.Results = results

            self.logger.info("sending write response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.BrowseRequest_Encoding_DefaultBinary):
            self.logger.info("Browse request")
            params = struct_from_binary(ua.BrowseParameters, body)

            results = self.session.browse(params)

            response = ua.BrowseResponse()
            response.Results = results

            self.logger.info("sending browse response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.GetEndpointsRequest_Encoding_DefaultBinary):
            self.logger.info("get endpoints request")
            params = struct_from_binary(ua.GetEndpointsParameters, body)

            endpoints = self.iserver.get_endpoints(params, sockname=self.sockname)

            response = ua.GetEndpointsResponse()
            response.Endpoints = endpoints

            self.logger.info("sending get endpoints response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.FindServersRequest_Encoding_DefaultBinary):
            self.logger.info("find servers request")
            params = struct_from_binary(ua.FindServersParameters, body)

            servers = self.local_discovery_service.find_servers(params)

            response = ua.FindServersResponse()
            response.Servers = servers

            self.logger.info("sending find servers response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.RegisterServerRequest_Encoding_DefaultBinary):
            self.logger.info("register server request")
            serv = struct_from_binary(ua.RegisteredServer, body)

            self.local_discovery_service.register_server(serv)

            response = ua.RegisterServerResponse()

            self.logger.info("sending register server response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.RegisterServer2Request_Encoding_DefaultBinary):
            self.logger.info("register server 2 request")
            params = struct_from_binary(ua.RegisterServer2Parameters, body)

            results = self.local_discovery_service.register_server2(params)

            response = ua.RegisterServer2Response()
            response.ConfigurationResults = results

            self.logger.info("sending register server 2 response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.TranslateBrowsePathsToNodeIdsRequest_Encoding_DefaultBinary):
            self.logger.info("translate browsepaths to nodeids request")
            params = struct_from_binary(ua.TranslateBrowsePathsToNodeIdsParameters, body)

            paths = self.session.translate_browsepaths_to_nodeids(params.BrowsePaths)

            response = ua.TranslateBrowsePathsToNodeIdsResponse()
            response.Results = paths

            self.logger.info("sending translate browsepaths to nodeids response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.AddNodesRequest_Encoding_DefaultBinary):
            self.logger.info("add nodes request")
            params = struct_from_binary(ua.AddNodesParameters, body)

            results = self.session.add_nodes(params.NodesToAdd)

            response = ua.AddNodesResponse()
            response.Results = results

            self.logger.info("sending add node response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.DeleteNodesRequest_Encoding_DefaultBinary):
            self.logger.info("delete nodes request")
            params = struct_from_binary(ua.DeleteNodesParameters, body)

            results = self.session.delete_nodes(params)

            response = ua.DeleteNodesResponse()
            response.Results = results

            self.logger.info("sending delete node response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.AddReferencesRequest_Encoding_DefaultBinary):
            self.logger.info("add references request")
            params = struct_from_binary(ua.AddReferencesParameters, body)

            results = self.session.add_references(params.ReferencesToAdd)

            response = ua.AddReferencesResponse()
            response.Results = results

            self.logger.info("sending add references response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.DeleteReferencesRequest_Encoding_DefaultBinary):
            self.logger.info("delete references request")
            params = struct_from_binary(ua.DeleteReferencesParameters, body)

            results = self.session.delete_references(params.ReferencesToDelete)

            response = ua.DeleteReferencesResponse()
            response.Parameters.Results = results

            self.logger.info("sending delete references response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)


        elif typeid == ua.NodeId(ua.ObjectIds.CreateSubscriptionRequest_Encoding_DefaultBinary):
            self.logger.info("create subscription request")
            params = struct_from_binary(ua.CreateSubscriptionParameters, body)

            result = self.session.create_subscription(params, self.forward_publish_response)

            response = ua.CreateSubscriptionResponse()
            response.Parameters = result

            self.logger.info("sending create subscription response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.ModifySubscriptionRequest_Encoding_DefaultBinary):
            self.logger.info("modify subscription request")
            params = struct_from_binary(ua.ModifySubscriptionParameters, body)

            result = self.session.modify_subscription(params, self.forward_publish_response)

            response = ua.ModifySubscriptionResponse()
            response.Parameters = result

            self.logger.info("sending modify subscription response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.DeleteSubscriptionsRequest_Encoding_DefaultBinary):
            self.logger.info("delete subscriptions request")
            params = struct_from_binary(ua.DeleteSubscriptionsParameters, body)

            results = self.session.delete_subscriptions(params.SubscriptionIds)

            response = ua.DeleteSubscriptionsResponse()
            response.Results = results

            self.logger.info("sending delte subscription response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.CreateMonitoredItemsRequest_Encoding_DefaultBinary):
            self.logger.info("create monitored items request")
            params = struct_from_binary(ua.CreateMonitoredItemsParameters, body)
            results = self.session.create_monitored_items(params)

            response = ua.CreateMonitoredItemsResponse()
            response.Results = results

            self.logger.info("sending create monitored items response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.ModifyMonitoredItemsRequest_Encoding_DefaultBinary):
            self.logger.info("modify monitored items request")
            params = struct_from_binary(ua.ModifyMonitoredItemsParameters, body)
            results = self.session.modify_monitored_items(params)

            response = ua.ModifyMonitoredItemsResponse()
            response.Results = results

            self.logger.info("sending modify monitored items response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.DeleteMonitoredItemsRequest_Encoding_DefaultBinary):
            self.logger.info("delete monitored items request")
            params = struct_from_binary(ua.DeleteMonitoredItemsParameters, body)

            results = self.session.delete_monitored_items(params)

            response = ua.DeleteMonitoredItemsResponse()
            response.Results = results

            self.logger.info("sending delete monitored items response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.HistoryReadRequest_Encoding_DefaultBinary):
            self.logger.info("history read request")
            params = struct_from_binary(ua.HistoryReadParameters, body)

            results = self.session.history_read(params)

            response = ua.HistoryReadResponse()
            response.Results = results

            self.logger.info("sending history read response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.RegisterNodesRequest_Encoding_DefaultBinary):
            self.logger.info("register nodes request")
            params = struct_from_binary(ua.RegisterNodesParameters, body)
            self.logger.info("Node registration not implemented")

            response = ua.RegisterNodesResponse()
            response.Parameters.RegisteredNodeIds = params.NodesToRegister

            self.logger.info("sending register nodes response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.UnregisterNodesRequest_Encoding_DefaultBinary):
            self.logger.info("unregister nodes request")
            params = struct_from_binary(ua.UnregisterNodesParameters, body)

            response = ua.UnregisterNodesResponse()

            self.logger.info("sending unregister nodes response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.PublishRequest_Encoding_DefaultBinary):
            self.logger.info("publish request")

            if not self.session:
                return False

            params = struct_from_binary(ua.PublishParameters, body)

            data = PublishRequestData()
            data.requesthdr = requesthdr
            data.seqhdr = seqhdr
            with self._datalock:
                self._publishdata_queue.append(data)  # will be used to send publish answers from server
                if self._publish_result_queue:
                    result = self._publish_result_queue.pop(0)
                    self.forward_publish_response(result)
            self.session.publish(params.SubscriptionAcknowledgements)
            self.logger.info("publish forward to server")

        elif typeid == ua.NodeId(ua.ObjectIds.RepublishRequest_Encoding_DefaultBinary):
            self.logger.info("re-publish request")

            params = struct_from_binary(ua.RepublishParameters, body)
            msg = self.session.republish(params)

            response = ua.RepublishResponse()
            response.NotificationMessage = msg

            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.CloseSecureChannelRequest_Encoding_DefaultBinary):
            self.logger.info("close secure channel request")
            self._connection.close()
            response = ua.CloseSecureChannelResponse()
            self.send_response(requesthdr.RequestHandle, seqhdr, response)
            return False

        elif typeid == ua.NodeId(ua.ObjectIds.CallRequest_Encoding_DefaultBinary):
            self.logger.info("call request")

            params = struct_from_binary(ua.CallParameters, body)

            results = self.session.call(params.MethodsToCall)

            response = ua.CallResponse()
            response.Results = results

            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.SetMonitoringModeRequest_Encoding_DefaultBinary):
            self.logger.info("set monitoring mode request")

            params = struct_from_binary(ua.SetMonitoringModeParameters, body)

            # FIXME: Implement SetMonitoringMode
            # Send dummy results to keep clients happy
            response = ua.SetMonitoringModeResponse()
            results = ua.SetMonitoringModeResult()
            ids = params.MonitoredItemIds
            statuses = [ua.StatusCode(ua.StatusCodes.Good) for node_id in ids]
            results.Results = statuses
            response.Parameters = results

            self.logger.info("sending set monitoring mode response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        elif typeid == ua.NodeId(ua.ObjectIds.SetPublishingModeRequest_Encoding_DefaultBinary):
            self.logger.info("set publishing mode request")

            params = struct_from_binary(ua.SetPublishingModeParameters, body)

            # FIXME: Implement SetPublishingMode
            # Send dummy results to keep clients happy
            response = ua.SetPublishingModeResponse()
            results = ua.SetPublishingModeResult()
            ids = params.SubscriptionIds
            statuses = [ua.StatusCode(ua.StatusCodes.Good) for node_id in ids]
            results.Results = statuses
            response.Parameters = results

            self.logger.info("sending set publishing mode response")
            self.send_response(requesthdr.RequestHandle, seqhdr, response)

        else:
            self.logger.warning("Unknown message received %s", typeid)
            raise utils.ServiceError(ua.StatusCodes.BadServiceUnsupported)

        return True

    def close(self):
        """
        to be called when client has disconnected to ensure we really close
        everything we should
        """
        self.logger.info("Cleanup client connection: %s", self.name)
        if self.session:
            self.session.close_session(True)



"""
Internal server implementing opcu-ua interface.
Can be used on server side or to implement binary/https opc-ua servers
"""
from datetime import datetime, timedelta
from copy import copy
import os
import logging
from threading import Lock
from enum import Enum
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


from opcua import ua
from opcua.common import utils
from opcua.common.callback import (CallbackType, ServerItemCallback,
                                   CallbackDispatcher)
from opcua.common.node import Node
from opcua.server.history import HistoryManager
from opcua.server.address_space import AddressSpace
from opcua.server.address_space import AttributeService
from opcua.server.address_space import ViewService
from opcua.server.address_space import NodeManagementService
from opcua.server.address_space import MethodService
from opcua.server.subscription_service import SubscriptionService
from opcua.server.discovery_service import LocalDiscoveryService
from opcua.server.standard_address_space import standard_address_space
from opcua.server.user_manager import UserManager
#from opcua.common import xmlimporter


class SessionState(Enum):
    Created = 0
    Activated = 1
    Closed = 2

class InternalServer(object):

    def __init__(self, shelffile=None, user_manager=None, session_cls=None):
        self.logger = logging.getLogger(__name__)

        self.server_callback_dispatcher = CallbackDispatcher()

        self.endpoints = []
        self._channel_id_counter = 5
        self.disabled_clock = False  # for debugging we may want to disable clock that writes too much in log
        self._local_discovery_service = None # lazy-loading

        self.aspace = AddressSpace()
        self.attribute_service = AttributeService(self.aspace)
        self.view_service = ViewService(self.aspace)
        self.method_service = MethodService(self.aspace)
        self.node_mgt_service = NodeManagementService(self.aspace)

        self.load_standard_address_space(shelffile)

        self.loop = None
        self.asyncio_transports = []
        self.subscription_service = SubscriptionService(self.aspace)

        self.history_manager = HistoryManager(self)
        self.user_manager = user_manager

        # create a session to use on server side
        self.session_cls = session_cls or InternalSession
        self.isession = self.session_cls(self, self.aspace, \
          self.subscription_service, "Internal", user=UserManager.User.Admin)

        self.server_status_node = Node(self.isession, ua.NodeId(ua.ObjectIds.Server_ServerStatus))
        self.current_time_node = Node(self.isession, ua.NodeId(ua.ObjectIds.Server_ServerStatus_CurrentTime))
        self._address_space_fixes()
        self.setup_nodes()

    @property
    def thread_loop(self):
        if self.loop is None:
            raise Exception("InternalServer stopped: async threadloop is not running.")
        return self.loop

    @property
    def local_discovery_service(self):
        if self._local_discovery_service is None:
            self._local_discovery_service = LocalDiscoveryService(parent = self)
            for edp in self.endpoints:
                srvDesc = LocalDiscoveryService.ServerDescription(edp.Server)
                self._local_discovery_service.add_server_description(srvDesc)
        return self._local_discovery_service

    def setup_nodes(self):
        """
        Set up some nodes as defined by spec
        """
        uries = ["http://opcfoundation.org/UA/"]
        ns_node = Node(self.isession, ua.NodeId(ua.ObjectIds.Server_NamespaceArray))
        ns_node.set_value(uries)

    def load_standard_address_space(self, shelffile=None):
        if (shelffile is not None) and (os.path.isfile(shelffile) or os.path.isfile(shelffile+".db")):
            # import address space from shelf
            self.aspace.load_aspace_shelf(shelffile)
        else:
            # import address space from code generated from xml
            standard_address_space.fill_address_space(self.node_mgt_service)
            # import address space directly from xml, this has performance impact so disabled
            # importer = xmlimporter.XmlImporter(self.node_mgt_service)
            # importer.import_xml("/path/to/python-opcua/schemas/Opc.Ua.NodeSet2.xml", self)

            # if a cache file was supplied a shelve of the standard address space can now be built for next start up
            if shelffile:
                self.aspace.make_aspace_shelf(shelffile)

    def _address_space_fixes(self):
        """
        Looks like the xml definition of address space has some error. This is a good place to fix them
        """

        it = ua.AddReferencesItem()
        it.SourceNodeId = ua.NodeId(ua.ObjectIds.BaseObjectType)
        it.ReferenceTypeId = ua.NodeId(ua.ObjectIds.Organizes)
        it.IsForward = False
        it.TargetNodeId = ua.NodeId(ua.ObjectIds.ObjectTypesFolder)
        it.TargetNodeClass = ua.NodeClass.Object

        it2 = ua.AddReferencesItem()
        it2.SourceNodeId = ua.NodeId(ua.ObjectIds.BaseDataType)
        it2.ReferenceTypeId = ua.NodeId(ua.ObjectIds.Organizes)
        it2.IsForward = False
        it2.TargetNodeId = ua.NodeId(ua.ObjectIds.DataTypesFolder)
        it2.TargetNodeClass = ua.NodeClass.Object

        results = self.isession.add_references([it, it2])

        params = ua.WriteParameters()
        for nodeid in (ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerRead,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadData,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryReadEvents,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerWrite,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateData,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerHistoryUpdateEvents,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerMethodCall,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerBrowse,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerRegisterNodes,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerTranslateBrowsePathsToNodeIds,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxNodesPerNodeManagement,
                     ua.ObjectIds.Server_ServerCapabilities_OperationLimits_MaxMonitoredItemsPerCall):
            attr = ua.WriteValue()
            attr.NodeId = ua.NodeId(nodeid)
            attr.AttributeId = ua.AttributeIds.Value
            attr.Value = ua.DataValue(ua.Variant(10000, ua.VariantType.UInt32), ua.StatusCode(ua.StatusCodes.Good))
            attr.Value.ServerTimestamp = datetime.utcnow()
            params.NodesToWrite.append(attr)
        result = self.isession.write(params)
        result[0].check()


    def load_address_space(self, path):
        """
        Load address space from path
        """
        self.aspace.load(path)

    def dump_address_space(self, path):
        """
        Dump current address space to path
        """
        self.aspace.dump(path)

    def start(self):
        self.logger.info("starting internal server")
        self.loop = utils.ThreadLoop()
        self.loop.start()
        self.subscription_service.set_loop(self.loop)
        serverState = Node(self.isession, ua.NodeId(ua.ObjectIds.Server_ServerStatus_State))
        serverState.set_value(ua.uaprotocol_auto.ServerState.Running, ua.VariantType.Int32)
        Node(self.isession, ua.NodeId(ua.ObjectIds.Server_ServerStatus_StartTime)).set_value(datetime.utcnow())
        if not self.disabled_clock:
            self._set_current_time()

    def stop(self):
        self.logger.info("stopping internal server")
        self.isession.close_session()
        self.subscription_service.set_loop(None)
        self.history_manager.stop()
        if self.loop:
            self.loop.stop()
            # wait for ThreadLoop to finish before proceeding
            self.loop.join()
            self.loop.close()
            self.loop = None

    def is_running(self):
        return self.loop is not None

    def _set_current_time(self):
        self.current_time_node.set_value(datetime.utcnow())
        ssdata = self.server_status_node.get_value()
        ssdata.CurrentTime = datetime.utcnow()
        self.server_status_node.set_value(ssdata)
        self.loop.call_later(1, self._set_current_time)

    def get_new_channel_id(self):
        self._channel_id_counter += 1
        return self._channel_id_counter

    def add_endpoint(self, endpoint):
        self.endpoints.append(endpoint)

    def get_endpoints(self, params=None, sockname=None):
        self.logger.info("get endpoint")
        if sockname:
            # return to client the ip address it has access to
            edps = []
            for edp in self.endpoints:
                edp1 = copy(edp)
                url = urlparse(edp1.EndpointUrl)
                url = url._replace(netloc=sockname[0] + ":" + str(sockname[1]))
                edp1.EndpointUrl = url.geturl()
                edps.append(edp1)
            return edps
        return self.endpoints[:]

    def create_session(self, name, user=UserManager.User.Anonymous):
        return self.session_cls(self, self.aspace, self.subscription_service, name, user=user)

    def enable_history_data_change(self, node, period=timedelta(days=7), count=0):
        """
        Set attribute Historizing of node to True and start storing data for history
        """
        node.set_attribute(ua.AttributeIds.Historizing, ua.DataValue(True))
        node.set_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.HistoryRead)
        node.set_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.HistoryRead)
        self.history_manager.historize_data_change(node, period, count)

    def disable_history_data_change(self, node):
        """
        Set attribute Historizing of node to False and stop storing data for history
        """
        node.set_attribute(ua.AttributeIds.Historizing, ua.DataValue(False))
        node.unset_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.HistoryRead)
        node.unset_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.HistoryRead)
        self.history_manager.dehistorize(node)

    def enable_history_event(self, source, period=timedelta(days=7), count=0):
        """
        Set attribute History Read of object events to True and start storing data for history
        """
        event_notifier = source.get_event_notifier()
        if ua.EventNotifier.SubscribeToEvents not in event_notifier:
            raise ua.UaError("Node does not generate events", event_notifier)

        if ua.EventNotifier.HistoryRead not in event_notifier:
            event_notifier.add(ua.EventNotifier.HistoryRead)
            source.set_event_notifier(event_notifier)

        self.history_manager.historize_event(source, period, count)

    def disable_history_event(self, source):
        """
        Set attribute History Read of node to False and stop storing data for history
        """
        source.unset_attr_bit(ua.AttributeIds.EventNotifier, ua.EventNotifier.HistoryRead)
        self.history_manager.dehistorize(source)

    def subscribe_server_callback(self, event, handle):
        """
        Create a subscription from event to handle
        """
        self.server_callback_dispatcher.addListener(event, handle)

    def unsubscribe_server_callback(self, event, handle):
        """
        Remove a subscription from event to handle
        """
        self.server_callback_dispatcher.removeListener(event, handle)

    def set_attribute_value(self, nodeid, datavalue, attr=ua.AttributeIds.Value):
        """
        directly write datavalue to the Attribute, bypasing some checks and structure creation
        so it is a little faster
        """
        self.aspace.set_attribute_value(nodeid, ua.AttributeIds.Value, datavalue)


class InternalSession(object):
    _counter = 10
    _auth_counter = 1000

    def __init__(self, internal_server, aspace, submgr, name, user=UserManager.User.Anonymous):
        self.logger = logging.getLogger(__name__)
        self.iserver = internal_server
        self.aspace = aspace
        self.subscription_service = submgr
        self.name = name
        self.user = user
        self.nonce = None
        self.state = SessionState.Created
        self.session_id = ua.NodeId(self._counter)
        InternalSession._counter += 1
        self.authentication_token = ua.NodeId(self._auth_counter)
        InternalSession._auth_counter += 1
        self.subscriptions = []
        self.logger.info("Created internal session %s", self.name)
        self._lock = Lock()

    @property
    def user_manager(self):
        return self.iserver.user_manager

    def __str__(self):
        return "InternalSession(name:{0}, user:{1}, id:{2}, auth_token:{3})".format(
            self.name, self.user, self.session_id, self.authentication_token)

    def get_endpoints(self, params=None, sockname=None):
        return self.iserver.get_endpoints(params, sockname)

    def create_session(self, params, sockname=None):
        self.logger.info("Create session request")

        result = ua.CreateSessionResult()
        result.SessionId = self.session_id
        result.AuthenticationToken = self.authentication_token
        result.RevisedSessionTimeout = params.RequestedSessionTimeout
        result.MaxRequestMessageSize = 65536
        self.nonce = utils.create_nonce(32)
        result.ServerNonce = self.nonce
        result.ServerEndpoints = self.get_endpoints(sockname=sockname)

        return result

    def close_session(self, delete_subs=True):
        self.logger.info("close session %s with subscriptions %s", self, self.subscriptions)
        self.state = SessionState.Closed
        self.delete_subscriptions(self.subscriptions[:])

    def activate_session(self, params):
        self.logger.info("activate session")
        result = ua.ActivateSessionResult()
        if self.state != SessionState.Created:
            raise utils.ServiceError(ua.StatusCodes.BadSessionIdInvalid)
        self.nonce = utils.create_nonce(32)
        result.ServerNonce = self.nonce
        for _ in params.ClientSoftwareCertificates:
            result.Results.append(ua.StatusCode())
        self.state = SessionState.Activated
        id_token = params.UserIdentityToken
        if isinstance(id_token, ua.UserNameIdentityToken):
            if self.user_manager.check_user_token(self, id_token) == False:
                raise utils.ServiceError(ua.StatusCodes.BadUserAccessDenied)
        self.logger.info("Activated internal session %s for user %s", self.name, self.user)
        return result

    def read(self, params):
        results = self.iserver.attribute_service.read(params)
        return results

    def history_read(self, params):
        return self.iserver.history_manager.read_history(params)

    def write(self, params):
        return self.iserver.attribute_service.write(params, self.user)

    def browse(self, params):
        return self.iserver.view_service.browse(params)

    def translate_browsepaths_to_nodeids(self, params):
        return self.iserver.view_service.translate_browsepaths_to_nodeids(params)

    def add_nodes(self, params):
        return self.iserver.node_mgt_service.add_nodes(params, self.user)

    def delete_nodes(self, params):
        return self.iserver.node_mgt_service.delete_nodes(params, self.user)

    def add_references(self, params):
        return self.iserver.node_mgt_service.add_references(params, self.user)

    def delete_references(self, params):
        return self.iserver.node_mgt_service.delete_references(params, self.user)

    def add_method_callback(self, methodid, callback):
        return self.aspace.add_method_callback(methodid, callback)

    def call(self, params):
        return self.iserver.method_service.call(params)

    def create_subscription(self, params, callback, ready_callback=None):
        result = self.subscription_service.create_subscription(params, callback)
        with self._lock:
            self.subscriptions.append(result.SubscriptionId)
        return result

    def modify_subscription(self, params, callback):
        return self.subscription_service.modify_subscription(params, callback)

    def create_monitored_items(self, params):
        subscription_result = self.subscription_service.create_monitored_items(params)
        self.iserver.server_callback_dispatcher.dispatch(
            CallbackType.ItemSubscriptionCreated, ServerItemCallback(params, subscription_result))
        return subscription_result

    def modify_monitored_items(self, params):
        subscription_result = self.subscription_service.modify_monitored_items(params)
        self.iserver.server_callback_dispatcher.dispatch(
            CallbackType.ItemSubscriptionModified, ServerItemCallback(params, subscription_result))
        return subscription_result

    def republish(self, params):
        return self.subscription_service.republish(params)

    def delete_subscriptions(self, ids):
        for i in ids:
            with self._lock:
                if i in self.subscriptions:
                    self.subscriptions.remove(i)
        return self.subscription_service.delete_subscriptions(ids)

    def delete_monitored_items(self, params):
        subscription_result = self.subscription_service.delete_monitored_items(params)
        self.iserver.server_callback_dispatcher.dispatch(
            CallbackType.ItemSubscriptionDeleted, ServerItemCallback(params, subscription_result))
        return subscription_result

    def publish(self, acks=None):
        if acks is None:
            acks = []
        return self.subscription_service.publish(acks)



print("import: 'opcua'")
import opcua

print("import: 'opcua.client'")
import opcua.client

print("import: 'opcua.common'")
import opcua.common

print("import: 'opcua.crypto'")
import opcua.crypto

print("import: 'opcua.server'")
import opcua.server

print("import: 'opcua.ua'")
import opcua.ua




"""
Helper function and classes that do not rely on opcua library.
Helper function and classes depending on ua object are in ua_utils.py
"""

import logging
import os
from concurrent.futures import Future
import functools
import threading
from socket import error as SocketError

try:
    import asyncio
except ImportError:
    import trollius as asyncio


from opcua.ua.uaerrors import UaError


class ServiceError(UaError):
    def __init__(self, code):
        super(ServiceError, self).__init__('UA Service Error')
        self.code = code


class NotEnoughData(UaError):
    pass


class SocketClosedException(UaError):
    pass


class Buffer(object):

    """
    alternative to io.BytesIO making debug easier
    and added a few conveniance methods
    """

    def __init__(self, data, start_pos=0, size=-1):
        # self.logger = logging.getLogger(__name__)
        self._data = data
        self._cur_pos = start_pos
        if size == -1:
            size = len(data) - start_pos
        self._size = size

    def __str__(self):
        return "Buffer(size:{0}, data:{1})".format(
            self._size,
            self._data[self._cur_pos:self._cur_pos + self._size])
    __repr__ = __str__

    def __len__(self):
        return self._size

    def read(self, size):
        """
        read and pop number of bytes for buffer
        """
        if size > self._size:
            raise NotEnoughData("Not enough data left in buffer, request for {0}, we have {1}".format(size, self))
        # self.logger.debug("Request for %s bytes, from %s", size, self)
        self._size -= size
        pos = self._cur_pos
        self._cur_pos += size
        data = self._data[pos:self._cur_pos]
        # self.logger.debug("Returning: %s ", data)
        return data

    def copy(self, size=-1):
        """
        return a shadow copy, optionnaly only copy 'size' bytes
        """
        if size == -1 or size > self._size:
            size = self._size
        return Buffer(self._data, self._cur_pos, size)

    def skip(self, size):
        """
        skip size bytes in buffer
        """
        if size > self._size:
            raise NotEnoughData("Not enough data left in buffer, request for {0}, we have {1}".format(size, self))
        self._size -= size
        self._cur_pos += size


class SocketWrapper(object):
    """
    wrapper to make it possible to have same api for
    normal sockets, socket from asyncio, StringIO, etc....
    """

    def __init__(self, sock):
        self.socket = sock

    def read(self, size):
        """
        Receive up to size bytes from socket
        """
        data = b''
        while size > 0:
            try:
                chunk = self.socket.recv(size)
            except (OSError, SocketError) as ex:
                raise SocketClosedException("Server socket has closed", ex)
            if not chunk:
                raise SocketClosedException("Server socket has closed")
            data += chunk
            size -= len(chunk)
        return data

    def write(self, data):
        self.socket.sendall(data)


def create_nonce(size=32):
    return os.urandom(size)


class ThreadLoop(threading.Thread):
    """
    run an asyncio loop in a thread
    """

    def __init__(self):
        threading.Thread.__init__(self)
        self.logger = logging.getLogger(__name__)
        self.loop = None
        self._cond = threading.Condition()

    def start(self):
        with self._cond:
            threading.Thread.start(self)
            self._cond.wait()

    def run(self):
        self.logger.debug("Starting subscription thread")
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        with self._cond:
            self._cond.notify_all()
        self.loop.run_forever()
        self.logger.debug("subscription thread ended")

    def create_server(self, proto, hostname, port):
        return self.loop.create_server(proto, hostname, port)

    def stop(self):
        """
        stop subscription loop, thus the subscription thread
        """
        self.loop.call_soon_threadsafe(self.loop.stop)

    def close(self):
        self.loop.close()
        self.loop = None

    def call_soon(self, callback):
        self.loop.call_soon_threadsafe(callback)

    def call_later(self, delay, callback):
        """
        threadsafe call_later from asyncio
        """
        p = functools.partial(self.loop.call_later, delay, callback)
        self.loop.call_soon_threadsafe(p)

    def _create_task(self, future, coro, cb=None):
        #task = self.loop.create_task(coro)
        task = asyncio.ensure_future(coro, loop=self.loop) 
        if cb:
            task.add_done_callback(cb)
        future.set_result(task)

    def create_task(self, coro, cb=None):
        """
        threadsafe create_task from asyncio
        """
        future = Future()
        p = functools.partial(self._create_task, future, coro, cb)
        self.loop.call_soon_threadsafe(p)
        return future.result()

    def run_coro_and_wait(self, coro):
        cond = threading.Condition()
        def cb(_):
            with cond:
                cond.notify_all()
        with cond:
            task = self.create_task(coro, cb)
            cond.wait()
        return task.result()

    def _run_until_complete(self, future, coro):
        task = self.loop.run_until_complete(coro)
        future.set_result(task)

    def run_until_complete(self, coro):
        """
        threadsafe run_until_completed from asyncio
        """
        future = Future()
        p = functools.partial(self._run_until_complete, future, coro)
        self.loop.call_soon_threadsafe(p)
        return future.result()






#AUTOGENERATED!!! Date: 2020-06-19 19:44:10.693270

from opcua.ua.uaerrors import UaStatusCodeError

class StatusCodes:
    Good = 0
    Uncertain = 0x40000000
    Bad = 0x80000000
    BadUnexpectedError = 0x80010000
    BadInternalError = 0x80020000
    BadOutOfMemory = 0x80030000
    BadResourceUnavailable = 0x80040000
    BadCommunicationError = 0x80050000
    BadEncodingError = 0x80060000
    BadDecodingError = 0x80070000
    BadEncodingLimitsExceeded = 0x80080000
    BadRequestTooLarge = 0x80B80000
    BadResponseTooLarge = 0x80B90000
    BadUnknownResponse = 0x80090000
    BadTimeout = 0x800A0000
    BadServiceUnsupported = 0x800B0000
    BadShutdown = 0x800C0000
    BadServerNotConnected = 0x800D0000
    BadServerHalted = 0x800E0000
    BadNothingToDo = 0x800F0000
    BadTooManyOperations = 0x80100000
    BadTooManyMonitoredItems = 0x80DB0000
    BadDataTypeIdUnknown = 0x80110000
    BadCertificateInvalid = 0x80120000
    BadSecurityChecksFailed = 0x80130000
    BadCertificatePolicyCheckFailed = 0x81140000
    BadCertificateTimeInvalid = 0x80140000
    BadCertificateIssuerTimeInvalid = 0x80150000
    BadCertificateHostNameInvalid = 0x80160000
    BadCertificateUriInvalid = 0x80170000
    BadCertificateUseNotAllowed = 0x80180000
    BadCertificateIssuerUseNotAllowed = 0x80190000
    BadCertificateUntrusted = 0x801A0000
    BadCertificateRevocationUnknown = 0x801B0000
    BadCertificateIssuerRevocationUnknown = 0x801C0000
    BadCertificateRevoked = 0x801D0000
    BadCertificateIssuerRevoked = 0x801E0000
    BadCertificateChainIncomplete = 0x810D0000
    BadUserAccessDenied = 0x801F0000
    BadIdentityTokenInvalid = 0x80200000
    BadIdentityTokenRejected = 0x80210000
    BadSecureChannelIdInvalid = 0x80220000
    BadInvalidTimestamp = 0x80230000
    BadNonceInvalid = 0x80240000
    BadSessionIdInvalid = 0x80250000
    BadSessionClosed = 0x80260000
    BadSessionNotActivated = 0x80270000
    BadSubscriptionIdInvalid = 0x80280000
    BadRequestHeaderInvalid = 0x802A0000
    BadTimestampsToReturnInvalid = 0x802B0000
    BadRequestCancelledByClient = 0x802C0000
    BadTooManyArguments = 0x80E50000
    BadLicenseExpired = 0x810E0000
    BadLicenseLimitsExceeded = 0x810F0000
    BadLicenseNotAvailable = 0x81100000
    GoodSubscriptionTransferred = 0x002D0000
    GoodCompletesAsynchronously = 0x002E0000
    GoodOverload = 0x002F0000
    GoodClamped = 0x00300000
    BadNoCommunication = 0x80310000
    BadWaitingForInitialData = 0x80320000
    BadNodeIdInvalid = 0x80330000
    BadNodeIdUnknown = 0x80340000
    BadAttributeIdInvalid = 0x80350000
    BadIndexRangeInvalid = 0x80360000
    BadIndexRangeNoData = 0x80370000
    BadDataEncodingInvalid = 0x80380000
    BadDataEncodingUnsupported = 0x80390000
    BadNotReadable = 0x803A0000
    BadNotWritable = 0x803B0000
    BadOutOfRange = 0x803C0000
    BadNotSupported = 0x803D0000
    BadNotFound = 0x803E0000
    BadObjectDeleted = 0x803F0000
    BadNotImplemented = 0x80400000
    BadMonitoringModeInvalid = 0x80410000
    BadMonitoredItemIdInvalid = 0x80420000
    BadMonitoredItemFilterInvalid = 0x80430000
    BadMonitoredItemFilterUnsupported = 0x80440000
    BadFilterNotAllowed = 0x80450000
    BadStructureMissing = 0x80460000
    BadEventFilterInvalid = 0x80470000
    BadContentFilterInvalid = 0x80480000
    BadFilterOperatorInvalid = 0x80C10000
    BadFilterOperatorUnsupported = 0x80C20000
    BadFilterOperandCountMismatch = 0x80C30000
    BadFilterOperandInvalid = 0x80490000
    BadFilterElementInvalid = 0x80C40000
    BadFilterLiteralInvalid = 0x80C50000
    BadContinuationPointInvalid = 0x804A0000
    BadNoContinuationPoints = 0x804B0000
    BadReferenceTypeIdInvalid = 0x804C0000
    BadBrowseDirectionInvalid = 0x804D0000
    BadNodeNotInView = 0x804E0000
    BadNumericOverflow = 0x81120000
    BadServerUriInvalid = 0x804F0000
    BadServerNameMissing = 0x80500000
    BadDiscoveryUrlMissing = 0x80510000
    BadSempahoreFileMissing = 0x80520000
    BadRequestTypeInvalid = 0x80530000
    BadSecurityModeRejected = 0x80540000
    BadSecurityPolicyRejected = 0x80550000
    BadTooManySessions = 0x80560000
    BadUserSignatureInvalid = 0x80570000
    BadApplicationSignatureInvalid = 0x80580000
    BadNoValidCertificates = 0x80590000
    BadIdentityChangeNotSupported = 0x80C60000
    BadRequestCancelledByRequest = 0x805A0000
    BadParentNodeIdInvalid = 0x805B0000
    BadReferenceNotAllowed = 0x805C0000
    BadNodeIdRejected = 0x805D0000
    BadNodeIdExists = 0x805E0000
    BadNodeClassInvalid = 0x805F0000
    BadBrowseNameInvalid = 0x80600000
    BadBrowseNameDuplicated = 0x80610000
    BadNodeAttributesInvalid = 0x80620000
    BadTypeDefinitionInvalid = 0x80630000
    BadSourceNodeIdInvalid = 0x80640000
    BadTargetNodeIdInvalid = 0x80650000
    BadDuplicateReferenceNotAllowed = 0x80660000
    BadInvalidSelfReference = 0x80670000
    BadReferenceLocalOnly = 0x80680000
    BadNoDeleteRights = 0x80690000
    UncertainReferenceNotDeleted = 0x40BC0000
    BadServerIndexInvalid = 0x806A0000
    BadViewIdUnknown = 0x806B0000
    BadViewTimestampInvalid = 0x80C90000
    BadViewParameterMismatch = 0x80CA0000
    BadViewVersionInvalid = 0x80CB0000
    UncertainNotAllNodesAvailable = 0x40C00000
    GoodResultsMayBeIncomplete = 0x00BA0000
    BadNotTypeDefinition = 0x80C80000
    UncertainReferenceOutOfServer = 0x406C0000
    BadTooManyMatches = 0x806D0000
    BadQueryTooComplex = 0x806E0000
    BadNoMatch = 0x806F0000
    BadMaxAgeInvalid = 0x80700000
    BadSecurityModeInsufficient = 0x80E60000
    BadHistoryOperationInvalid = 0x80710000
    BadHistoryOperationUnsupported = 0x80720000
    BadInvalidTimestampArgument = 0x80BD0000
    BadWriteNotSupported = 0x80730000
    BadTypeMismatch = 0x80740000
    BadMethodInvalid = 0x80750000
    BadArgumentsMissing = 0x80760000
    BadNotExecutable = 0x81110000
    BadTooManySubscriptions = 0x80770000
    BadTooManyPublishRequests = 0x80780000
    BadNoSubscription = 0x80790000
    BadSequenceNumberUnknown = 0x807A0000
    BadMessageNotAvailable = 0x807B0000
    BadInsufficientClientProfile = 0x807C0000
    BadStateNotActive = 0x80BF0000
    BadAlreadyExists = 0x81150000
    BadTcpServerTooBusy = 0x807D0000
    BadTcpMessageTypeInvalid = 0x807E0000
    BadTcpSecureChannelUnknown = 0x807F0000
    BadTcpMessageTooLarge = 0x80800000
    BadTcpNotEnoughResources = 0x80810000
    BadTcpInternalError = 0x80820000
    BadTcpEndpointUrlInvalid = 0x80830000
    BadRequestInterrupted = 0x80840000
    BadRequestTimeout = 0x80850000
    BadSecureChannelClosed = 0x80860000
    BadSecureChannelTokenUnknown = 0x80870000
    BadSequenceNumberInvalid = 0x80880000
    BadProtocolVersionUnsupported = 0x80BE0000
    BadConfigurationError = 0x80890000
    BadNotConnected = 0x808A0000
    BadDeviceFailure = 0x808B0000
    BadSensorFailure = 0x808C0000
    BadOutOfService = 0x808D0000
    BadDeadbandFilterInvalid = 0x808E0000
    UncertainNoCommunicationLastUsableValue = 0x408F0000
    UncertainLastUsableValue = 0x40900000
    UncertainSubstituteValue = 0x40910000
    UncertainInitialValue = 0x40920000
    UncertainSensorNotAccurate = 0x40930000
    UncertainEngineeringUnitsExceeded = 0x40940000
    UncertainSubNormal = 0x40950000
    GoodLocalOverride = 0x00960000
    BadRefreshInProgress = 0x80970000
    BadConditionAlreadyDisabled = 0x80980000
    BadConditionAlreadyEnabled = 0x80CC0000
    BadConditionDisabled = 0x80990000
    BadEventIdUnknown = 0x809A0000
    BadEventNotAcknowledgeable = 0x80BB0000
    BadDialogNotActive = 0x80CD0000
    BadDialogResponseInvalid = 0x80CE0000
    BadConditionBranchAlreadyAcked = 0x80CF0000
    BadConditionBranchAlreadyConfirmed = 0x80D00000
    BadConditionAlreadyShelved = 0x80D10000
    BadConditionNotShelved = 0x80D20000
    BadShelvingTimeOutOfRange = 0x80D30000
    BadNoData = 0x809B0000
    BadBoundNotFound = 0x80D70000
    BadBoundNotSupported = 0x80D80000
    BadDataLost = 0x809D0000
    BadDataUnavailable = 0x809E0000
    BadEntryExists = 0x809F0000
    BadNoEntryExists = 0x80A00000
    BadTimestampNotSupported = 0x80A10000
    GoodEntryInserted = 0x00A20000
    GoodEntryReplaced = 0x00A30000
    UncertainDataSubNormal = 0x40A40000
    GoodNoData = 0x00A50000
    GoodMoreData = 0x00A60000
    BadAggregateListMismatch = 0x80D40000
    BadAggregateNotSupported = 0x80D50000
    BadAggregateInvalidInputs = 0x80D60000
    BadAggregateConfigurationRejected = 0x80DA0000
    GoodDataIgnored = 0x00D90000
    BadRequestNotAllowed = 0x80E40000
    BadRequestNotComplete = 0x81130000
    GoodEdited = 0x00DC0000
    GoodPostActionFailed = 0x00DD0000
    UncertainDominantValueChanged = 0x40DE0000
    GoodDependentValueChanged = 0x00E00000
    BadDominantValueChanged = 0x80E10000
    UncertainDependentValueChanged = 0x40E20000
    BadDependentValueChanged = 0x80E30000
    GoodEdited_DependentValueChanged = 0x01160000
    GoodEdited_DominantValueChanged = 0x01170000
    GoodEdited_DominantValueChanged_DependentValueChanged = 0x01180000
    BadEdited_OutOfRange = 0x81190000
    BadInitialValue_OutOfRange = 0x811A0000
    BadOutOfRange_DominantValueChanged = 0x811B0000
    BadEdited_OutOfRange_DominantValueChanged = 0x811C0000
    BadOutOfRange_DominantValueChanged_DependentValueChanged = 0x811D0000
    BadEdited_OutOfRange_DominantValueChanged_DependentValueChanged = 0x811E0000
    GoodCommunicationEvent = 0x00A70000
    GoodShutdownEvent = 0x00A80000
    GoodCallAgain = 0x00A90000
    GoodNonCriticalTimeout = 0x00AA0000
    BadInvalidArgument = 0x80AB0000
    BadConnectionRejected = 0x80AC0000
    BadDisconnect = 0x80AD0000
    BadConnectionClosed = 0x80AE0000
    BadInvalidState = 0x80AF0000
    BadEndOfStream = 0x80B00000
    BadNoDataAvailable = 0x80B10000
    BadWaitingForResponse = 0x80B20000
    BadOperationAbandoned = 0x80B30000
    BadExpectedStreamToBlock = 0x80B40000
    BadWouldBlock = 0x80B50000
    BadSyntaxError = 0x80B60000
    BadMaxConnectionsReached = 0x80B70000


code_to_name_doc = {
    0: ('Good', 'The operation completed successfully.'),
    0x40000000: ('Uncertain', 'The operation completed however its outputs may not be usable.'),
    0x80000000: ('Bad', 'The operation failed.'),
    0x80010000: ('BadUnexpectedError', '"An unexpected error occurred."'),
    0x80020000: ('BadInternalError', '"An internal error occurred as a result of a programming or configuration error."'),
    0x80030000: ('BadOutOfMemory', '"Not enough memory to complete the operation."'),
    0x80040000: ('BadResourceUnavailable', '"An operating system resource is not available."'),
    0x80050000: ('BadCommunicationError', '"A low level communication error occurred."'),
    0x80060000: ('BadEncodingError', '"Encoding halted because of invalid data in the objects being serialized."'),
    0x80070000: ('BadDecodingError', '"Decoding halted because of invalid data in the stream."'),
    0x80080000: ('BadEncodingLimitsExceeded', '"The message encoding/decoding limits imposed by the stack have been exceeded."'),
    0x80B80000: ('BadRequestTooLarge', '"The request message size exceeds limits set by the server."'),
    0x80B90000: ('BadResponseTooLarge', '"The response message size exceeds limits set by the client."'),
    0x80090000: ('BadUnknownResponse', '"An unrecognized response was received from the server."'),
    0x800A0000: ('BadTimeout', '"The operation timed out."'),
    0x800B0000: ('BadServiceUnsupported', '"The server does not support the requested service."'),
    0x800C0000: ('BadShutdown', '"The operation was cancelled because the application is shutting down."'),
    0x800D0000: ('BadServerNotConnected', '"The operation could not complete because the client is not connected to the server."'),
    0x800E0000: ('BadServerHalted', '"The server has stopped and cannot process any requests."'),
    0x800F0000: ('BadNothingToDo', '"There was nothing to do because the client passed a list of operations with no elements."'),
    0x80100000: ('BadTooManyOperations', '"The request could not be processed because it specified too many operations."'),
    0x80DB0000: ('BadTooManyMonitoredItems', '"The request could not be processed because there are too many monitored items in the subscription."'),
    0x80110000: ('BadDataTypeIdUnknown', '"The extension object cannot be (de)serialized because the data type id is not recognized."'),
    0x80120000: ('BadCertificateInvalid', '"The certificate provided as a parameter is not valid."'),
    0x80130000: ('BadSecurityChecksFailed', '"An error occurred verifying security."'),
    0x81140000: ('BadCertificatePolicyCheckFailed', '"The certificate does not meet the requirements of the security policy."'),
    0x80140000: ('BadCertificateTimeInvalid', '"The certificate has expired or is not yet valid."'),
    0x80150000: ('BadCertificateIssuerTimeInvalid', '"An issuer certificate has expired or is not yet valid."'),
    0x80160000: ('BadCertificateHostNameInvalid', '"The HostName used to connect to a server does not match a HostName in the certificate."'),
    0x80170000: ('BadCertificateUriInvalid', '"The URI specified in the ApplicationDescription does not match the URI in the certificate."'),
    0x80180000: ('BadCertificateUseNotAllowed', '"The certificate may not be used for the requested operation."'),
    0x80190000: ('BadCertificateIssuerUseNotAllowed', '"The issuer certificate may not be used for the requested operation."'),
    0x801A0000: ('BadCertificateUntrusted', '"The certificate is not trusted."'),
    0x801B0000: ('BadCertificateRevocationUnknown', '"It was not possible to determine if the certificate has been revoked."'),
    0x801C0000: ('BadCertificateIssuerRevocationUnknown', '"It was not possible to determine if the issuer certificate has been revoked."'),
    0x801D0000: ('BadCertificateRevoked', '"The certificate has been revoked."'),
    0x801E0000: ('BadCertificateIssuerRevoked', '"The issuer certificate has been revoked."'),
    0x810D0000: ('BadCertificateChainIncomplete', '"The certificate chain is incomplete."'),
    0x801F0000: ('BadUserAccessDenied', '"User does not have permission to perform the requested operation."'),
    0x80200000: ('BadIdentityTokenInvalid', '"The user identity token is not valid."'),
    0x80210000: ('BadIdentityTokenRejected', '"The user identity token is valid but the server has rejected it."'),
    0x80220000: ('BadSecureChannelIdInvalid', '"The specified secure channel is no longer valid."'),
    0x80230000: ('BadInvalidTimestamp', '"The timestamp is outside the range allowed by the server."'),
    0x80240000: ('BadNonceInvalid', '"The nonce does appear to be not a random value or it is not the correct length."'),
    0x80250000: ('BadSessionIdInvalid', '"The session id is not valid."'),
    0x80260000: ('BadSessionClosed', '"The session was closed by the client."'),
    0x80270000: ('BadSessionNotActivated', '"The session cannot be used because ActivateSession has not been called."'),
    0x80280000: ('BadSubscriptionIdInvalid', '"The subscription id is not valid."'),
    0x802A0000: ('BadRequestHeaderInvalid', '"The header for the request is missing or invalid."'),
    0x802B0000: ('BadTimestampsToReturnInvalid', '"The timestamps to return parameter is invalid."'),
    0x802C0000: ('BadRequestCancelledByClient', '"The request was cancelled by the client."'),
    0x80E50000: ('BadTooManyArguments', '"Too many arguments were provided."'),
    0x810E0000: ('BadLicenseExpired', '"The server requires a license to operate in general or to perform a service or operation, but existing license is expired."'),
    0x810F0000: ('BadLicenseLimitsExceeded', '"The server has limits on number of allowed operations / objects, based on installed licenses, and these limits where exceeded."'),
    0x81100000: ('BadLicenseNotAvailable', '"The server does not have a license which is required to operate in general or to perform a service or operation."'),
    0x002D0000: ('GoodSubscriptionTransferred', '"The subscription was transferred to another session."'),
    0x002E0000: ('GoodCompletesAsynchronously', '"The processing will complete asynchronously."'),
    0x002F0000: ('GoodOverload', '"Sampling has slowed down due to resource limitations."'),
    0x00300000: ('GoodClamped', '"The value written was accepted but was clamped."'),
    0x80310000: ('BadNoCommunication', '"Communication with the data source is defined, but not established, and there is no last known value available."'),
    0x80320000: ('BadWaitingForInitialData', '"Waiting for the server to obtain values from the underlying data source."'),
    0x80330000: ('BadNodeIdInvalid', '"The syntax of the node id is not valid."'),
    0x80340000: ('BadNodeIdUnknown', '"The node id refers to a node that does not exist in the server address space."'),
    0x80350000: ('BadAttributeIdInvalid', '"The attribute is not supported for the specified Node."'),
    0x80360000: ('BadIndexRangeInvalid', '"The syntax of the index range parameter is invalid."'),
    0x80370000: ('BadIndexRangeNoData', '"No data exists within the range of indexes specified."'),
    0x80380000: ('BadDataEncodingInvalid', '"The data encoding is invalid."'),
    0x80390000: ('BadDataEncodingUnsupported', '"The server does not support the requested data encoding for the node."'),
    0x803A0000: ('BadNotReadable', '"The access level does not allow reading or subscribing to the Node."'),
    0x803B0000: ('BadNotWritable', '"The access level does not allow writing to the Node."'),
    0x803C0000: ('BadOutOfRange', '"The value was out of range."'),
    0x803D0000: ('BadNotSupported', '"The requested operation is not supported."'),
    0x803E0000: ('BadNotFound', '"A requested item was not found or a search operation ended without success."'),
    0x803F0000: ('BadObjectDeleted', '"The object cannot be used because it has been deleted."'),
    0x80400000: ('BadNotImplemented', '"Requested operation is not implemented."'),
    0x80410000: ('BadMonitoringModeInvalid', '"The monitoring mode is invalid."'),
    0x80420000: ('BadMonitoredItemIdInvalid', '"The monitoring item id does not refer to a valid monitored item."'),
    0x80430000: ('BadMonitoredItemFilterInvalid', '"The monitored item filter parameter is not valid."'),
    0x80440000: ('BadMonitoredItemFilterUnsupported', '"The server does not support the requested monitored item filter."'),
    0x80450000: ('BadFilterNotAllowed', '"A monitoring filter cannot be used in combination with the attribute specified."'),
    0x80460000: ('BadStructureMissing', '"A mandatory structured parameter was missing or null."'),
    0x80470000: ('BadEventFilterInvalid', '"The event filter is not valid."'),
    0x80480000: ('BadContentFilterInvalid', '"The content filter is not valid."'),
    0x80C10000: ('BadFilterOperatorInvalid', '"An unrecognized operator was provided in a filter."'),
    0x80C20000: ('BadFilterOperatorUnsupported', '"A valid operator was provided, but the server does not provide support for this filter operator."'),
    0x80C30000: ('BadFilterOperandCountMismatch', '"The number of operands provided for the filter operator was less then expected for the operand provided."'),
    0x80490000: ('BadFilterOperandInvalid', '"The operand used in a content filter is not valid."'),
    0x80C40000: ('BadFilterElementInvalid', '"The referenced element is not a valid element in the content filter."'),
    0x80C50000: ('BadFilterLiteralInvalid', '"The referenced literal is not a valid value."'),
    0x804A0000: ('BadContinuationPointInvalid', '"The continuation point provide is longer valid."'),
    0x804B0000: ('BadNoContinuationPoints', '"The operation could not be processed because all continuation points have been allocated."'),
    0x804C0000: ('BadReferenceTypeIdInvalid', '"The reference type id does not refer to a valid reference type node."'),
    0x804D0000: ('BadBrowseDirectionInvalid', '"The browse direction is not valid."'),
    0x804E0000: ('BadNodeNotInView', '"The node is not part of the view."'),
    0x81120000: ('BadNumericOverflow', '"The number was not accepted because of a numeric overflow."'),
    0x804F0000: ('BadServerUriInvalid', '"The ServerUri is not a valid URI."'),
    0x80500000: ('BadServerNameMissing', '"No ServerName was specified."'),
    0x80510000: ('BadDiscoveryUrlMissing', '"No DiscoveryUrl was specified."'),
    0x80520000: ('BadSempahoreFileMissing', '"The semaphore file specified by the client is not valid."'),
    0x80530000: ('BadRequestTypeInvalid', '"The security token request type is not valid."'),
    0x80540000: ('BadSecurityModeRejected', '"The security mode does not meet the requirements set by the server."'),
    0x80550000: ('BadSecurityPolicyRejected', '"The security policy does not meet the requirements set by the server."'),
    0x80560000: ('BadTooManySessions', '"The server has reached its maximum number of sessions."'),
    0x80570000: ('BadUserSignatureInvalid', '"The user token signature is missing or invalid."'),
    0x80580000: ('BadApplicationSignatureInvalid', '"The signature generated with the client certificate is missing or invalid."'),
    0x80590000: ('BadNoValidCertificates', '"The client did not provide at least one software certificate that is valid and meets the profile requirements for the server."'),
    0x80C60000: ('BadIdentityChangeNotSupported', '"The server does not support changing the user identity assigned to the session."'),
    0x805A0000: ('BadRequestCancelledByRequest', '"The request was cancelled by the client with the Cancel service."'),
    0x805B0000: ('BadParentNodeIdInvalid', '"The parent node id does not to refer to a valid node."'),
    0x805C0000: ('BadReferenceNotAllowed', '"The reference could not be created because it violates constraints imposed by the data model."'),
    0x805D0000: ('BadNodeIdRejected', '"The requested node id was reject because it was either invalid or server does not allow node ids to be specified by the client."'),
    0x805E0000: ('BadNodeIdExists', '"The requested node id is already used by another node."'),
    0x805F0000: ('BadNodeClassInvalid', '"The node class is not valid."'),
    0x80600000: ('BadBrowseNameInvalid', '"The browse name is invalid."'),
    0x80610000: ('BadBrowseNameDuplicated', '"The browse name is not unique among nodes that share the same relationship with the parent."'),
    0x80620000: ('BadNodeAttributesInvalid', '"The node attributes are not valid for the node class."'),
    0x80630000: ('BadTypeDefinitionInvalid', '"The type definition node id does not reference an appropriate type node."'),
    0x80640000: ('BadSourceNodeIdInvalid', '"The source node id does not reference a valid node."'),
    0x80650000: ('BadTargetNodeIdInvalid', '"The target node id does not reference a valid node."'),
    0x80660000: ('BadDuplicateReferenceNotAllowed', '"The reference type between the nodes is already defined."'),
    0x80670000: ('BadInvalidSelfReference', '"The server does not allow this type of self reference on this node."'),
    0x80680000: ('BadReferenceLocalOnly', '"The reference type is not valid for a reference to a remote server."'),
    0x80690000: ('BadNoDeleteRights', '"The server will not allow the node to be deleted."'),
    0x40BC0000: ('UncertainReferenceNotDeleted', '"The server was not able to delete all target references."'),
    0x806A0000: ('BadServerIndexInvalid', '"The server index is not valid."'),
    0x806B0000: ('BadViewIdUnknown', '"The view id does not refer to a valid view node."'),
    0x80C90000: ('BadViewTimestampInvalid', '"The view timestamp is not available or not supported."'),
    0x80CA0000: ('BadViewParameterMismatch', '"The view parameters are not consistent with each other."'),
    0x80CB0000: ('BadViewVersionInvalid', '"The view version is not available or not supported."'),
    0x40C00000: ('UncertainNotAllNodesAvailable', '"The list of references may not be complete because the underlying system is not available."'),
    0x00BA0000: ('GoodResultsMayBeIncomplete', '"The server should have followed a reference to a node in a remote server but did not. The result set may be incomplete."'),
    0x80C80000: ('BadNotTypeDefinition', '"The provided Nodeid was not a type definition nodeid."'),
    0x406C0000: ('UncertainReferenceOutOfServer', '"One of the references to follow in the relative path references to a node in the address space in another server."'),
    0x806D0000: ('BadTooManyMatches', '"The requested operation has too many matches to return."'),
    0x806E0000: ('BadQueryTooComplex', '"The requested operation requires too many resources in the server."'),
    0x806F0000: ('BadNoMatch', '"The requested operation has no match to return."'),
    0x80700000: ('BadMaxAgeInvalid', '"The max age parameter is invalid."'),
    0x80E60000: ('BadSecurityModeInsufficient', '"The operation is not permitted over the current secure channel."'),
    0x80710000: ('BadHistoryOperationInvalid', '"The history details parameter is not valid."'),
    0x80720000: ('BadHistoryOperationUnsupported', '"The server does not support the requested operation."'),
    0x80BD0000: ('BadInvalidTimestampArgument', '"The defined timestamp to return was invalid."'),
    0x80730000: ('BadWriteNotSupported', '"The server does not support writing the combination of value, status and timestamps provided."'),
    0x80740000: ('BadTypeMismatch', '"The value supplied for the attribute is not of the same type as the attribute"s value."'),
    0x80750000: ('BadMethodInvalid', '"The method id does not refer to a method for the specified object."'),
    0x80760000: ('BadArgumentsMissing', '"The client did not specify all of the input arguments for the method."'),
    0x81110000: ('BadNotExecutable', '"The executable attribute does not allow the execution of the method."'),
    0x80770000: ('BadTooManySubscriptions', '"The server has reached its maximum number of subscriptions."'),
    0x80780000: ('BadTooManyPublishRequests', '"The server has reached the maximum number of queued publish requests."'),
    0x80790000: ('BadNoSubscription', '"There is no subscription available for this session."'),
    0x807A0000: ('BadSequenceNumberUnknown', '"The sequence number is unknown to the server."'),
    0x807B0000: ('BadMessageNotAvailable', '"The requested notification message is no longer available."'),
    0x807C0000: ('BadInsufficientClientProfile', '"The client of the current session does not support one or more Profiles that are necessary for the subscription."'),
    0x80BF0000: ('BadStateNotActive', '"The sub-state machine is not currently active."'),
    0x81150000: ('BadAlreadyExists', '"An equivalent rule already exists."'),
    0x807D0000: ('BadTcpServerTooBusy', '"The server cannot process the request because it is too busy."'),
    0x807E0000: ('BadTcpMessageTypeInvalid', '"The type of the message specified in the header invalid."'),
    0x807F0000: ('BadTcpSecureChannelUnknown', '"The SecureChannelId and/or TokenId are not currently in use."'),
    0x80800000: ('BadTcpMessageTooLarge', '"The size of the message specified in the header is too large."'),
    0x80810000: ('BadTcpNotEnoughResources', '"There are not enough resources to process the request."'),
    0x80820000: ('BadTcpInternalError', '"An internal error occurred."'),
    0x80830000: ('BadTcpEndpointUrlInvalid', '"The server does not recognize the QueryString specified."'),
    0x80840000: ('BadRequestInterrupted', '"The request could not be sent because of a network interruption."'),
    0x80850000: ('BadRequestTimeout', '"Timeout occurred while processing the request."'),
    0x80860000: ('BadSecureChannelClosed', '"The secure channel has been closed."'),
    0x80870000: ('BadSecureChannelTokenUnknown', '"The token has expired or is not recognized."'),
    0x80880000: ('BadSequenceNumberInvalid', '"The sequence number is not valid."'),
    0x80BE0000: ('BadProtocolVersionUnsupported', '"The applications do not have compatible protocol versions."'),
    0x80890000: ('BadConfigurationError', '"There is a problem with the configuration that affects the usefulness of the value."'),
    0x808A0000: ('BadNotConnected', '"The variable should receive its value from another variable, but has never been configured to do so."'),
    0x808B0000: ('BadDeviceFailure', '"There has been a failure in the device/data source that generates the value that has affected the value."'),
    0x808C0000: ('BadSensorFailure', '"There has been a failure in the sensor from which the value is derived by the device/data source."'),
    0x808D0000: ('BadOutOfService', '"The source of the data is not operational."'),
    0x808E0000: ('BadDeadbandFilterInvalid', '"The deadband filter is not valid."'),
    0x408F0000: ('UncertainNoCommunicationLastUsableValue', '"Communication to the data source has failed. The variable value is the last value that had a good quality."'),
    0x40900000: ('UncertainLastUsableValue', '"Whatever was updating this value has stopped doing so."'),
    0x40910000: ('UncertainSubstituteValue', '"The value is an operational value that was manually overwritten."'),
    0x40920000: ('UncertainInitialValue', '"The value is an initial value for a variable that normally receives its value from another variable."'),
    0x40930000: ('UncertainSensorNotAccurate', '"The value is at one of the sensor limits."'),
    0x40940000: ('UncertainEngineeringUnitsExceeded', '"The value is outside of the range of values defined for this parameter."'),
    0x40950000: ('UncertainSubNormal', '"The value is derived from multiple sources and has less than the required number of Good sources."'),
    0x00960000: ('GoodLocalOverride', '"The value has been overridden."'),
    0x80970000: ('BadRefreshInProgress', '"This Condition refresh failed, a Condition refresh operation is already in progress."'),
    0x80980000: ('BadConditionAlreadyDisabled', '"This condition has already been disabled."'),
    0x80CC0000: ('BadConditionAlreadyEnabled', '"This condition has already been enabled."'),
    0x80990000: ('BadConditionDisabled', '"Property not available, this condition is disabled."'),
    0x809A0000: ('BadEventIdUnknown', '"The specified event id is not recognized."'),
    0x80BB0000: ('BadEventNotAcknowledgeable', '"The event cannot be acknowledged."'),
    0x80CD0000: ('BadDialogNotActive', '"The dialog condition is not active."'),
    0x80CE0000: ('BadDialogResponseInvalid', '"The response is not valid for the dialog."'),
    0x80CF0000: ('BadConditionBranchAlreadyAcked', '"The condition branch has already been acknowledged."'),
    0x80D00000: ('BadConditionBranchAlreadyConfirmed', '"The condition branch has already been confirmed."'),
    0x80D10000: ('BadConditionAlreadyShelved', '"The condition has already been shelved."'),
    0x80D20000: ('BadConditionNotShelved', '"The condition is not currently shelved."'),
    0x80D30000: ('BadShelvingTimeOutOfRange', '"The shelving time not within an acceptable range."'),
    0x809B0000: ('BadNoData', '"No data exists for the requested time range or event filter."'),
    0x80D70000: ('BadBoundNotFound', '"No data found to provide upper or lower bound value."'),
    0x80D80000: ('BadBoundNotSupported', '"The server cannot retrieve a bound for the variable."'),
    0x809D0000: ('BadDataLost', '"Data is missing due to collection started/stopped/lost."'),
    0x809E0000: ('BadDataUnavailable', '"Expected data is unavailable for the requested time range due to an un-mounted volume, an off-line archive or tape, or similar reason for temporary unavailability."'),
    0x809F0000: ('BadEntryExists', '"The data or event was not successfully inserted because a matching entry exists."'),
    0x80A00000: ('BadNoEntryExists', '"The data or event was not successfully updated because no matching entry exists."'),
    0x80A10000: ('BadTimestampNotSupported', '"The client requested history using a timestamp format the server does not support (i.e requested ServerTimestamp when server only supports SourceTimestamp)."'),
    0x00A20000: ('GoodEntryInserted', '"The data or event was successfully inserted into the historical database."'),
    0x00A30000: ('GoodEntryReplaced', '"The data or event field was successfully replaced in the historical database."'),
    0x40A40000: ('UncertainDataSubNormal', '"The value is derived from multiple values and has less than the required number of Good values."'),
    0x00A50000: ('GoodNoData', '"No data exists for the requested time range or event filter."'),
    0x00A60000: ('GoodMoreData', '"The data or event field was successfully replaced in the historical database."'),
    0x80D40000: ('BadAggregateListMismatch', '"The requested number of Aggregates does not match the requested number of NodeIds."'),
    0x80D50000: ('BadAggregateNotSupported', '"The requested Aggregate is not support by the server."'),
    0x80D60000: ('BadAggregateInvalidInputs', '"The aggregate value could not be derived due to invalid data inputs."'),
    0x80DA0000: ('BadAggregateConfigurationRejected', '"The aggregate configuration is not valid for specified node."'),
    0x00D90000: ('GoodDataIgnored', '"The request specifies fields which are not valid for the EventType or cannot be saved by the historian."'),
    0x80E40000: ('BadRequestNotAllowed', '"The request was rejected by the server because it did not meet the criteria set by the server."'),
    0x81130000: ('BadRequestNotComplete', '"The request has not been processed by the server yet."'),
    0x00DC0000: ('GoodEdited', '"The value does not come from the real source and has been edited by the server."'),
    0x00DD0000: ('GoodPostActionFailed', '"There was an error in execution of these post-actions."'),
    0x40DE0000: ('UncertainDominantValueChanged', '"The related EngineeringUnit has been changed but the Variable Value is still provided based on the previous unit."'),
    0x00E00000: ('GoodDependentValueChanged', '"A dependent value has been changed but the change has not been applied to the device."'),
    0x80E10000: ('BadDominantValueChanged', '"The related EngineeringUnit has been changed but this change has not been applied to the device. The Variable Value is still dependent on the previous unit but its status is currently Bad."'),
    0x40E20000: ('UncertainDependentValueChanged', '"A dependent value has been changed but the change has not been applied to the device. The quality of the dominant variable is uncertain."'),
    0x80E30000: ('BadDependentValueChanged', '"A dependent value has been changed but the change has not been applied to the device. The quality of the dominant variable is Bad."'),
    0x01160000: ('GoodEdited_DependentValueChanged', '"It is delivered with a dominant Variable value when a dependent Variable has changed but the change has not been applied."'),
    0x01170000: ('GoodEdited_DominantValueChanged', '"It is delivered with a dependent Variable value when a dominant Variable has changed but the change has not been applied."'),
    0x01180000: ('GoodEdited_DominantValueChanged_DependentValueChanged', '"It is delivered with a dependent Variable value when a dominant or dependent Variable has changed but change has not been applied."'),
    0x81190000: ('BadEdited_OutOfRange', '"It is delivered with a Variable value when Variable has changed but the value is not legal."'),
    0x811A0000: ('BadInitialValue_OutOfRange', '"It is delivered with a Variable value when a source Variable has changed but the value is not legal."'),
    0x811B0000: ('BadOutOfRange_DominantValueChanged', '"It is delivered with a dependent Variable value when a dominant Variable has changed and the value is not legal."'),
    0x811C0000: ('BadEdited_OutOfRange_DominantValueChanged', '"It is delivered with a dependent Variable value when a dominant Variable has changed, the value is not legal and the change has not been applied."'),
    0x811D0000: ('BadOutOfRange_DominantValueChanged_DependentValueChanged', '"It is delivered with a dependent Variable value when a dominant or dependent Variable has changed and the value is not legal."'),
    0x811E0000: ('BadEdited_OutOfRange_DominantValueChanged_DependentValueChanged', '"It is delivered with a dependent Variable value when a dominant or dependent Variable has changed, the value is not legal and the change has not been applied."'),
    0x00A70000: ('GoodCommunicationEvent', '"The communication layer has raised an event."'),
    0x00A80000: ('GoodShutdownEvent', '"The system is shutting down."'),
    0x00A90000: ('GoodCallAgain', '"The operation is not finished and needs to be called again."'),
    0x00AA0000: ('GoodNonCriticalTimeout', '"A non-critical timeout occurred."'),
    0x80AB0000: ('BadInvalidArgument', '"One or more arguments are invalid."'),
    0x80AC0000: ('BadConnectionRejected', '"Could not establish a network connection to remote server."'),
    0x80AD0000: ('BadDisconnect', '"The server has disconnected from the client."'),
    0x80AE0000: ('BadConnectionClosed', '"The network connection has been closed."'),
    0x80AF0000: ('BadInvalidState', '"The operation cannot be completed because the object is closed, uninitialized or in some other invalid state."'),
    0x80B00000: ('BadEndOfStream', '"Cannot move beyond end of the stream."'),
    0x80B10000: ('BadNoDataAvailable', '"No data is currently available for reading from a non-blocking stream."'),
    0x80B20000: ('BadWaitingForResponse', '"The asynchronous operation is waiting for a response."'),
    0x80B30000: ('BadOperationAbandoned', '"The asynchronous operation was abandoned by the caller."'),
    0x80B40000: ('BadExpectedStreamToBlock', '"The stream did not return all data requested (possibly because it is a non-blocking stream)."'),
    0x80B50000: ('BadWouldBlock', '"Non blocking behaviour is required and the operation would block."'),
    0x80B60000: ('BadSyntaxError', '"A value had an invalid syntax."'),
    0x80B70000: ('BadMaxConnectionsReached', '"The operation could not be finished because all available connections are in use."'),
}


def get_name_and_doc(val):
        if val in code_to_name_doc:
            return code_to_name_doc[val]
        else:
            if val & 1 << 31:
                return 'Bad', 'Unknown StatusCode value: {}'.format(val)
            elif val & 1 << 30:
                return 'UncertainIn', 'Unknown StatusCode value: {}'.format(val)
            else:
                return 'Good', 'Unknown StatusCode value: {}'.format(val)
    


"""
from a list of nodes in the address space, build an XML file
format is the one from opc-ua specification
"""
import logging
from collections import OrderedDict
import xml.etree.ElementTree as Et
from copy import copy
import base64

from opcua import ua
from opcua.ua import object_ids as o_ids
from opcua.common.ua_utils import get_base_data_type
from opcua.ua.uatypes import extension_object_ids
from opcua.ua.uaerrors import UaError


class XmlExporter(object):

    ''' If it is required that for _extobj_to_etree members to the value should be written in a certain
        order it can be added to the dictionary below.
    '''
    extobj_ordered_elements = {
        ua.NodeId(ua.ObjectIds.Argument) : ['Name',
                                            'DataType',
                                            'ValueRank',
                                            'ArrayDimensions',
                                            'Description']
        }

    def __init__(self, server):
        self.logger = logging.getLogger(__name__)
        self.server = server
        self.aliases = {}
        self._addr_idx_to_xml_idx = {}

        node_set_attributes = OrderedDict()
        node_set_attributes['xmlns:xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
        node_set_attributes['xmlns:uax'] = 'http://opcfoundation.org/UA/2008/02/Types.xsd'
        node_set_attributes['xmlns:xsd'] = 'http://www.w3.org/2001/XMLSchema'
        node_set_attributes['xmlns'] = 'http://opcfoundation.org/UA/2011/03/UANodeSet.xsd'

        self.etree = Et.ElementTree(Et.Element('UANodeSet', node_set_attributes))

    def build_etree(self, node_list, uris=None):
        """
        Create an XML etree object from a list of nodes; custom namespace uris are optional
        Namespaces used by nodes are always exported for consistency.
        Args:
            node_list: list of Node objects for export
            uris: list of namespace uri strings

        Returns:
        """
        self.logger.info('Building XML etree')

        self._add_namespaces(node_list, uris)

        # add all nodes in the list to the XML etree
        for node in node_list:
            self.node_to_etree(node)

        # add aliases to the XML etree
        self._add_alias_els()

    def _add_namespaces(self, nodes, uris):
        idxs = self._get_ns_idxs_of_nodes(nodes)

        ns_array = self.server.get_namespace_array()

        # now add index of provided uris if necessary
        if uris:
            self._add_idxs_from_uris(idxs, uris, ns_array)

        # now create a dict of idx_in_address_space to idx_in_exported_file
        self._addr_idx_to_xml_idx = self._make_idx_dict(idxs, ns_array)
        ns_to_export = [ns_array[i] for i in sorted(list(self._addr_idx_to_xml_idx.keys())) if i != 0]

        # write namespaces to xml
        self._add_namespace_uri_els(ns_to_export)

    def _make_idx_dict(self, idxs, ns_array):
        idxs.sort()
        addr_idx_to_xml_idx = {0: 0}
        for xml_idx, addr_idx in enumerate(idxs):
            if addr_idx >= len(ns_array):
                break
            addr_idx_to_xml_idx[addr_idx] = xml_idx + 1
        return addr_idx_to_xml_idx

    def _get_ns_idxs_of_nodes(self, nodes):
        """
        get a list of all indexes used or references by nodes
        """
        idxs = []
        for node in nodes:
            node_idxs = [node.nodeid.NamespaceIndex]
            try:
                node_idxs.append(node.get_browse_name().NamespaceIndex)
            except UaError:
                self.logger.exception("Error retrieving browse name of node %s", node)
                raise

            node_idxs.extend(ref.NodeId.NamespaceIndex for ref in node.get_references())
            node_idxs = list(set(node_idxs))  # remove duplicates
            for i in node_idxs:
                if i != 0 and i not in idxs:
                    idxs.append(i)
        return idxs

    def _add_idxs_from_uris(self, idxs, uris, ns_array):
        for uri in uris:
            if uri in ns_array:
                i = ns_array.index(uri)
                if i not in idxs:
                    idxs.append(i)


    def write_xml(self, xmlpath, pretty=True):
        """
        Write the XML etree in the exporter object to a file
        Args:
            xmlpath: string representing the path/file name

        Returns:
        """
        # try to write the XML etree to a file
        self.logger.info('Exporting XML file to %s', xmlpath)
        # from IPython import embed
        # embed()
        if pretty:
            indent(self.etree.getroot())
            self.etree.write(xmlpath,
                             encoding='utf-8',
                             xml_declaration=True
                            )
        else:
            self.etree.write(xmlpath,
                             encoding='utf-8',
                             xml_declaration=True
                            )

    def dump_etree(self):
        """
        Dump etree to console for debugging
        Returns:
        """
        self.logger.info('Dumping XML etree to console')
        Et.dump(self.etree)

    def node_to_etree(self, node):
        """
        Add the necessary XML sub elements to the etree for exporting the node
        Args:
            node: Node object which will be added to XML etree

        Returns:
        """
        node_class = node.get_node_class()

        if node_class is ua.NodeClass.Object:
            self.add_etree_object(node)
        elif node_class is ua.NodeClass.ObjectType:
            self.add_etree_object_type(node)
        elif node_class is ua.NodeClass.Variable:
            self.add_etree_variable(node)
        elif node_class is ua.NodeClass.VariableType:
            self.add_etree_variable_type(node)
        elif node_class is ua.NodeClass.ReferenceType:
            self.add_etree_reference_type(node)
        elif node_class is ua.NodeClass.DataType:
            self.add_etree_datatype(node)
        elif node_class is ua.NodeClass.Method:
            self.add_etree_method(node)
        else:
            self.logger.info("Exporting node class not implemented: %s ", node_class)

    def _add_sub_el(self, el, name, text):
        child_el = Et.SubElement(el, name)
        child_el.text = text
        return child_el

    def _node_to_string(self, nodeid):
        if not isinstance(nodeid, ua.NodeId):
            nodeid = nodeid.nodeid

        if nodeid.NamespaceIndex in self._addr_idx_to_xml_idx:
            nodeid = copy(nodeid)
            nodeid.NamespaceIndex = self._addr_idx_to_xml_idx[nodeid.NamespaceIndex]
        return nodeid.to_string()

    def _bname_to_string(self, bname):
        if bname.NamespaceIndex in self._addr_idx_to_xml_idx:
            bname = copy(bname)
            bname.NamespaceIndex = self._addr_idx_to_xml_idx[bname.NamespaceIndex]
        return bname.to_string()

    def _add_node_common(self, nodetype, node):
        browsename = node.get_browse_name()
        nodeid = node.nodeid
        parent = node.get_parent()
        displayname = node.get_display_name().Text
        desc = node.get_description().Text if node.get_description() else None
        node_el = Et.SubElement(self.etree.getroot(), nodetype)
        node_el.attrib["NodeId"] = self._node_to_string(nodeid)
        node_el.attrib["BrowseName"] = self._bname_to_string(browsename)
        if parent is not None:
            node_class = node.get_node_class()
            if node_class in (ua.NodeClass.Object, ua.NodeClass.Variable, ua.NodeClass.Method):
                node_el.attrib["ParentNodeId"] = self._node_to_string(parent)
        self._add_sub_el(node_el, 'DisplayName', displayname)
        if desc not in (None, ""):
            self._add_sub_el(node_el, 'Description', desc)
        # FIXME: add WriteMask and UserWriteMask
        return node_el

    def add_etree_object(self, node):
        """
        Add a UA object element to the XML etree
        """
        obj_el = self._add_node_common("UAObject", node)
        var = node.get_attribute(ua.AttributeIds.EventNotifier)
        if var.Value.Value != 0:
            obj_el.attrib["EventNotifier"] = str(var.Value.Value)
        self._add_ref_els(obj_el, node)

    def add_etree_object_type(self, node):
        """
        Add a UA object type element to the XML etree
        """
        obj_el = self._add_node_common("UAObjectType", node)
        abstract = node.get_attribute(ua.AttributeIds.IsAbstract).Value.Value
        if abstract:
            obj_el.attrib["IsAbstract"] = 'true'
        self._add_ref_els(obj_el, node)

    def add_variable_common(self, node, el):
        dtype = node.get_data_type()
        if dtype.NamespaceIndex == 0 and dtype.Identifier in o_ids.ObjectIdNames:
            dtype_name = o_ids.ObjectIdNames[dtype.Identifier]
            self.aliases[dtype] = dtype_name
        else:
            dtype_name = self._node_to_string(dtype)
        rank = node.get_value_rank()
        if rank != -1:
            el.attrib["ValueRank"] = str(int(rank))
        dim = node.get_attribute(ua.AttributeIds.ArrayDimensions)
        if dim.Value.Value:
            el.attrib["ArrayDimensions"] = ",".join([str(i) for i in dim.Value.Value])
        el.attrib["DataType"] = dtype_name
        self.value_to_etree(el, dtype_name, dtype, node)

    def add_etree_variable(self, node):
        """
        Add a UA variable element to the XML etree
        """
        var_el = self._add_node_common("UAVariable", node)
        self._add_ref_els(var_el, node)
        self.add_variable_common(node, var_el)

        accesslevel = node.get_attribute(ua.AttributeIds.AccessLevel).Value.Value
        useraccesslevel = node.get_attribute(ua.AttributeIds.UserAccessLevel).Value.Value

        # We only write these values if they are different from defaults
        # Not sure where default is defined....
        if accesslevel not in (0, ua.AccessLevel.CurrentRead.mask):
            var_el.attrib["AccessLevel"] = str(accesslevel)
        if useraccesslevel not in (0, ua.AccessLevel.CurrentRead.mask):
            var_el.attrib["UserAccessLevel"] = str(useraccesslevel)

        var = node.get_attribute(ua.AttributeIds.MinimumSamplingInterval)
        if var.Value.Value:
            var_el.attrib["MinimumSamplingInterval"] = str(var.Value.Value)
        var = node.get_attribute(ua.AttributeIds.Historizing)
        if var.Value.Value:
            var_el.attrib["Historizing"] = 'true'

    def add_etree_variable_type(self, node):
        """
        Add a UA variable type element to the XML etree
        """

        var_el = self._add_node_common("UAVariableType", node)
        self.add_variable_common(node, var_el)

        abstract = node.get_attribute(ua.AttributeIds.IsAbstract)
        if abstract.Value.Value:
            var_el.attrib["IsAbstract"] = "true"

        self._add_ref_els(var_el, node)

    def add_etree_method(self, node):
        obj_el = self._add_node_common("UAMethod", node)

        var = node.get_attribute(ua.AttributeIds.Executable)
        if var.Value.Value is False:
            obj_el.attrib["Executable"] = "false"
        var = node.get_attribute(ua.AttributeIds.UserExecutable)
        if var.Value.Value is False:
            obj_el.attrib["UserExecutable"] = "false"
        self._add_ref_els(obj_el, node)

    def add_etree_reference_type(self, obj):
        obj_el = self._add_node_common("UAReferenceType", obj)
        self._add_ref_els(obj_el, obj)
        var = obj.get_attribute(ua.AttributeIds.InverseName)
        if var is not None and var.Value.Value is not None and var.Value.Value.Text is not None:
            self._add_sub_el(obj_el, 'InverseName', var.Value.Value.Text)

    def add_etree_datatype(self, obj):
        """
        Add a UA data type element to the XML etree
        """
        obj_el = self._add_node_common("UADataType", obj)
        self._add_ref_els(obj_el, obj)

    def _add_namespace_uri_els(self, uris):
        nuris_el = Et.Element('NamespaceUris')

        for uri in uris:
            self._add_sub_el(nuris_el, 'Uri', uri)

        self.etree.getroot().insert(0, nuris_el)

    def _add_alias_els(self):
        aliases_el = Et.Element('Aliases')

        ordered_keys = list(self.aliases.keys())
        ordered_keys.sort()
        for nodeid in ordered_keys:
            name = self.aliases[nodeid]
            ref_el = Et.SubElement(aliases_el, 'Alias', Alias=name)
            ref_el.text = self._node_to_string(nodeid)

        # insert behind the namespace element
        self.etree.getroot().insert(1, aliases_el)

    def _add_ref_els(self, parent_el, obj):
        refs = obj.get_references()
        refs_el = Et.SubElement(parent_el, 'References')

        for ref in refs:
            if ref.ReferenceTypeId.NamespaceIndex == 0 and ref.ReferenceTypeId.Identifier in o_ids.ObjectIdNames:
                ref_name = o_ids.ObjectIdNames[ref.ReferenceTypeId.Identifier]
            else:
                ref_name = self.server.get_node(ref.ReferenceTypeId).get_browse_name().Name

            ref_el = Et.SubElement(refs_el, 'Reference')
            ref_el.attrib['ReferenceType'] = ref_name
            if not ref.IsForward:
                ref_el.attrib['IsForward'] = 'false'
            ref_el.text = self._node_to_string(ref.NodeId)

            self.aliases[ref.ReferenceTypeId] = ref_name


    def member_to_etree(self, el, name, dtype, val):
        member_el = Et.SubElement(el, "uax:" + name)
        if isinstance(val, (list, tuple)):
            for v in val:
                self._value_to_etree(member_el, ua.ObjectIdNames[dtype.Identifier], dtype, v)
        else:
            self._val_to_etree(member_el, dtype, val)


    def _val_to_etree(self, el, dtype, val):
        if dtype == ua.NodeId(ua.ObjectIds.NodeId):
            id_el = Et.SubElement(el, "uax:Identifier")
            id_el.text = val.to_string()
        elif dtype == ua.NodeId(ua.ObjectIds.Guid):
            id_el = Et.SubElement(el, "uax:String")
            id_el.text = str(val)
        elif dtype == ua.NodeId(ua.ObjectIds.Boolean):
            el.text = 'true' if val else 'false'
        elif dtype == ua.NodeId(ua.ObjectIds.ByteString):
            if val is None:
                val = b""
            data = base64.b64encode(val)
            el.text = data.decode("utf-8")
        elif not hasattr(val, "ua_types"):
            if isinstance(val, bytes):
                # FIXME: should we also encode this (localized text I guess) using base64??
                el.text = val.decode("utf-8")
            else:
                if val is not None:
                    el.text = str(val)
        else:
            for name, vtype in val.ua_types:
                self.member_to_etree(el, name, ua.NodeId(getattr(ua.ObjectIds, vtype)), getattr(val, name))


    def value_to_etree(self, el, dtype_name, dtype, node):
        var = node.get_data_value().Value
        if var.Value is not None:
            val_el = Et.SubElement(el, 'Value')
            self._value_to_etree(val_el, dtype_name, dtype, var.Value)

    def _value_to_etree(self, el, type_name, dtype, val):
        if val is None:
            return

        if isinstance(val, (list, tuple)):
            if dtype.NamespaceIndex == 0 and dtype.Identifier <= 21:
                elname = "uax:ListOf" + type_name
            else:  # this is an extentionObject:
                elname = "uax:ListOfExtensionObject"

            list_el = Et.SubElement(el, elname)
            for nval in val:
                self._value_to_etree(list_el, type_name, dtype, nval)
        else:
            dtype_base = get_base_data_type(self.server.get_node(dtype))
            dtype_base = dtype_base.nodeid

            if dtype_base == ua.NodeId(ua.ObjectIds.Enumeration):
                dtype_base = ua.NodeId(ua.ObjectIds.Int32)
                type_name = ua.ObjectIdNames[dtype_base.Identifier]

            if dtype_base.NamespaceIndex == 0 and dtype_base.Identifier <= 21:
                type_name = ua.ObjectIdNames[dtype_base.Identifier]
                val_el = Et.SubElement(el, "uax:" + type_name)
                self._val_to_etree(val_el, dtype_base, val)
            else:
                self._extobj_to_etree(el, type_name, dtype, val)




    def _extobj_to_etree(self, val_el, name, dtype, val):
        obj_el = Et.SubElement(val_el, "uax:ExtensionObject")
        type_el = Et.SubElement(obj_el, "uax:TypeId")
        id_el = Et.SubElement(type_el, "uax:Identifier")
        id_el.text = dtype.to_string()
        body_el = Et.SubElement(obj_el, "uax:Body")
        struct_el = Et.SubElement(body_el, "uax:" + name)
        for name, vtype  in val.ua_types:
            # FIXME; what happend if we have a custom type which is not part of ObjectIds???
            if vtype.startswith("ListOf"):
                vtype = vtype[6:]
            self.member_to_etree(struct_el, name, ua.NodeId(getattr(ua.ObjectIds, vtype)), getattr(val, name))
            #self.member_to_etree(struct_el, name, extension_object_ids[vtype], getattr(val, name))
        #for name in self._get_member_order(dtype, val):
            #self.member_to_etree(struct_el, name, ua.NodeId(getattr(ua.ObjectIds, val.ua_types[name])), getattr(val, name))

    def _get_member_order(self, dtype, val):
        '''
        If an dtype has an entry in XmlExporter.extobj_ordered_elements return the export order of the elements
        else return the unordered members.
        '''
        if dtype not in XmlExporter.extobj_ordered_elements.keys():
            return val.ua_types.keys()
        else:
            member_keys = [name for name in XmlExporter.extobj_ordered_elements[dtype] if name in val.ua_types.keys() and getattr(val, name) is not None ]

        return member_keys

def indent(elem, level=0):
    """
    copy and paste from http://effbot.org/zone/element-lib.htm#prettyprint
    it basically walks your tree and adds spaces and newlines so the tree is
    printed in a nice way
    """
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i



"""
server side implementation of subscription service
"""

from threading import RLock
import logging

from opcua import ua
from opcua.common import utils
from opcua.server.internal_subscription import InternalSubscription


class SubscriptionService(object):

    def __init__(self, aspace):
        self.logger = logging.getLogger(__name__)
        self.loop = None
        self.aspace = aspace
        self.subscriptions = {}
        self._sub_id_counter = 77
        self._lock = RLock()

    def set_loop(self, loop):
        self.loop = loop

    def create_subscription(self, params, callback):
        self.logger.info("create subscription with callback: %s", callback)
        result = ua.CreateSubscriptionResult()
        result.RevisedPublishingInterval = params.RequestedPublishingInterval
        result.RevisedLifetimeCount = params.RequestedLifetimeCount
        result.RevisedMaxKeepAliveCount = params.RequestedMaxKeepAliveCount
        with self._lock:
            self._sub_id_counter += 1
            result.SubscriptionId = self._sub_id_counter

            sub = InternalSubscription(self, result, self.aspace, callback)
            sub.start()
            self.subscriptions[result.SubscriptionId] = sub

            return result

    def modify_subscription(self, params, callback):
        # Requested params are ignored, result = params set during create_subscription.
        self.logger.info("modify subscription with callback: %s", callback)
        result = ua.ModifySubscriptionResult()
        try:
            with self._lock:
                sub = self.subscriptions[params.SubscriptionId]
                result.RevisedPublishingInterval = sub.data.RevisedPublishingInterval
                result.RevisedLifetimeCount = sub.data.RevisedLifetimeCount
                result.RevisedMaxKeepAliveCount = sub.data.RevisedMaxKeepAliveCount

                return result
        except KeyError:
            raise utils.ServiceError(ua.StatusCodes.BadSubscriptionIdInvalid)

    def delete_subscriptions(self, ids):
        self.logger.info("delete subscriptions: %s", ids)
        res = []
        for i in ids:
            with self._lock:
                if i not in self.subscriptions:
                    res.append(ua.StatusCode(ua.StatusCodes.BadSubscriptionIdInvalid))
                else:
                    sub = self.subscriptions.pop(i)
                    sub.stop()
                    res.append(ua.StatusCode())
        return res

    def publish(self, acks):
        self.logger.info("publish request with acks %s", acks)
        with self._lock:
            for subid, sub in self.subscriptions.items():
                sub.publish([ack.SequenceNumber for ack in acks if ack.SubscriptionId == subid])

    def create_monitored_items(self, params):
        self.logger.info("create monitored items")
        with self._lock:
            if params.SubscriptionId not in self.subscriptions:
                res = []
                for _ in params.ItemsToCreate:
                    response = ua.MonitoredItemCreateResult()
                    response.StatusCode = ua.StatusCode(ua.StatusCodes.BadSubscriptionIdInvalid)
                    res.append(response)
                return res
            return self.subscriptions[params.SubscriptionId].monitored_item_srv.create_monitored_items(params)

    def modify_monitored_items(self, params):
        self.logger.info("modify monitored items")
        with self._lock:
            if params.SubscriptionId not in self.subscriptions:
                res = []
                for _ in params.ItemsToModify:
                    result = ua.MonitoredItemModifyResult()
                    result.StatusCode = ua.StatusCode(ua.StatusCodes.BadSubscriptionIdInvalid)
                    res.append(result)
                return res
            return self.subscriptions[params.SubscriptionId].monitored_item_srv.modify_monitored_items(params)

    def delete_monitored_items(self, params):
        self.logger.info("delete monitored items")
        with self._lock:
            if params.SubscriptionId not in self.subscriptions:
                res = []
                for _ in params.MonitoredItemIds:
                    res.append(ua.StatusCode(ua.StatusCodes.BadSubscriptionIdInvalid))
                return res
            return self.subscriptions[params.SubscriptionId].monitored_item_srv.delete_monitored_items(
                params.MonitoredItemIds)

    def republish(self, params):
        with self._lock:
            if params.SubscriptionId not in self.subscriptions:
                # TODO: what should I do?
                return ua.NotificationMessage()
            return self.subscriptions[params.SubscriptionId].republish(params.RetransmitSequenceNumber)

    def trigger_event(self, event):
        with self._lock:
            for sub in self.subscriptions.values():
                sub.monitored_item_srv.trigger_event(event)
